# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: label_selectors.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='label_selectors.proto',
  package='bentoml',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x15label_selectors.proto\x12\x07\x62\x65ntoml\"\x9c\x03\n\x0eLabelSelectors\x12>\n\x0cmatch_labels\x18\x01 \x03(\x0b\x32(.bentoml.LabelSelectors.MatchLabelsEntry\x12J\n\x11match_expressions\x18\x02 \x03(\x0b\x32/.bentoml.LabelSelectors.LabelSelectorExpression\x1a\xc9\x01\n\x17LabelSelectorExpression\x12O\n\x08operator\x18\x01 \x01(\x0e\x32=.bentoml.LabelSelectors.LabelSelectorExpression.OPERATOR_TYPE\x12\x0b\n\x03key\x18\x02 \x01(\t\x12\x0e\n\x06values\x18\x03 \x03(\t\"@\n\rOPERATOR_TYPE\x12\x06\n\x02In\x10\x00\x12\t\n\x05NotIn\x10\x01\x12\n\n\x06\x45xists\x10\x02\x12\x10\n\x0c\x44oesNotExist\x10\x03\x1a\x32\n\x10MatchLabelsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x62\x06proto3'
)



_LABELSELECTORS_LABELSELECTOREXPRESSION_OPERATOR_TYPE = _descriptor.EnumDescriptor(
  name='OPERATOR_TYPE',
  full_name='bentoml.LabelSelectors.LabelSelectorExpression.OPERATOR_TYPE',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='In', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='NotIn', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='Exists', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DoesNotExist', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=331,
  serialized_end=395,
)
_sym_db.RegisterEnumDescriptor(_LABELSELECTORS_LABELSELECTOREXPRESSION_OPERATOR_TYPE)


_LABELSELECTORS_LABELSELECTOREXPRESSION = _descriptor.Descriptor(
  name='LabelSelectorExpression',
  full_name='bentoml.LabelSelectors.LabelSelectorExpression',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='operator', full_name='bentoml.LabelSelectors.LabelSelectorExpression.operator', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='key', full_name='bentoml.LabelSelectors.LabelSelectorExpression.key', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='values', full_name='bentoml.LabelSelectors.LabelSelectorExpression.values', index=2,
      number=3, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _LABELSELECTORS_LABELSELECTOREXPRESSION_OPERATOR_TYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=194,
  serialized_end=395,
)

_LABELSELECTORS_MATCHLABELSENTRY = _descriptor.Descriptor(
  name='MatchLabelsEntry',
  full_name='bentoml.LabelSelectors.MatchLabelsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='bentoml.LabelSelectors.MatchLabelsEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='bentoml.LabelSelectors.MatchLabelsEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=397,
  serialized_end=447,
)

_LABELSELECTORS = _descriptor.Descriptor(
  name='LabelSelectors',
  full_name='bentoml.LabelSelectors',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='match_labels', full_name='bentoml.LabelSelectors.match_labels', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='match_expressions', full_name='bentoml.LabelSelectors.match_expressions', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_LABELSELECTORS_LABELSELECTOREXPRESSION, _LABELSELECTORS_MATCHLABELSENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=35,
  serialized_end=447,
)

_LABELSELECTORS_LABELSELECTOREXPRESSION.fields_by_name['operator'].enum_type = _LABELSELECTORS_LABELSELECTOREXPRESSION_OPERATOR_TYPE
_LABELSELECTORS_LABELSELECTOREXPRESSION.containing_type = _LABELSELECTORS
_LABELSELECTORS_LABELSELECTOREXPRESSION_OPERATOR_TYPE.containing_type = _LABELSELECTORS_LABELSELECTOREXPRESSION
_LABELSELECTORS_MATCHLABELSENTRY.containing_type = _LABELSELECTORS
_LABELSELECTORS.fields_by_name['match_labels'].message_type = _LABELSELECTORS_MATCHLABELSENTRY
_LABELSELECTORS.fields_by_name['match_expressions'].message_type = _LABELSELECTORS_LABELSELECTOREXPRESSION
DESCRIPTOR.message_types_by_name['LabelSelectors'] = _LABELSELECTORS
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

LabelSelectors = _reflection.GeneratedProtocolMessageType('LabelSelectors', (_message.Message,), {

  'LabelSelectorExpression' : _reflection.GeneratedProtocolMessageType('LabelSelectorExpression', (_message.Message,), {
    'DESCRIPTOR' : _LABELSELECTORS_LABELSELECTOREXPRESSION,
    '__module__' : 'label_selectors_pb2'
    # @@protoc_insertion_point(class_scope:bentoml.LabelSelectors.LabelSelectorExpression)
    })
  ,

  'MatchLabelsEntry' : _reflection.GeneratedProtocolMessageType('MatchLabelsEntry', (_message.Message,), {
    'DESCRIPTOR' : _LABELSELECTORS_MATCHLABELSENTRY,
    '__module__' : 'label_selectors_pb2'
    # @@protoc_insertion_point(class_scope:bentoml.LabelSelectors.MatchLabelsEntry)
    })
  ,
  'DESCRIPTOR' : _LABELSELECTORS,
  '__module__' : 'label_selectors_pb2'
  # @@protoc_insertion_point(class_scope:bentoml.LabelSelectors)
  })
_sym_db.RegisterMessage(LabelSelectors)
_sym_db.RegisterMessage(LabelSelectors.LabelSelectorExpression)
_sym_db.RegisterMessage(LabelSelectors.MatchLabelsEntry)


_LABELSELECTORS_MATCHLABELSENTRY._options = None
# @@protoc_insertion_point(module_scope)



import atexit
from concurrent import futures
import logging
import os
import subprocess
import time

import certifi
import click
from simple_di import Provide, inject

from bentoml.configuration import get_debug_mode
from bentoml.configuration.containers import BentoMLContainer
from bentoml.exceptions import BentoMLException
from bentoml.utils import reserve_free_port
from bentoml.yatai.utils import ensure_node_available_or_raise, parse_grpc_url


@inject
def get_yatai_service(
    channel_address: str = Provide[BentoMLContainer.config.yatai.remote.url],
    access_token: str = Provide[BentoMLContainer.config.yatai.remote.access_token],
    access_token_header: str = Provide[
        BentoMLContainer.config.yatai.remote.access_token_header
    ],
    tls_root_ca_cert: str = Provide[BentoMLContainer.yatai_tls_root_ca_cert],
    tls_client_key: str = Provide[BentoMLContainer.config.yatai.remote.tls.client_key],
    tls_client_cert: str = Provide[
        BentoMLContainer.config.yatai.remote.tls.client_cert
    ],
    db_url: str = Provide[BentoMLContainer.yatai_database_url],
    default_namespace: str = Provide[BentoMLContainer.config.yatai.namespace],
    repository_type: str = Provide[BentoMLContainer.config.yatai.repository.type],
    file_system_directory: str = Provide[BentoMLContainer.yatai_file_system_directory],
    s3_url: str = Provide[BentoMLContainer.config.yatai.repository.s3.url],
    s3_endpoint_url: str = Provide[
        BentoMLContainer.config.yatai.repository.s3.endpoint_url
    ],
    gcs_url: str = Provide[BentoMLContainer.config.yatai.repository.gcs.url],
):
    if channel_address:
        # Lazily import grpcio for YataiSerivce gRPC related actions
        import grpc
        from bentoml.yatai.client.interceptor import header_client_interceptor
        from bentoml.yatai.proto.yatai_service_pb2_grpc import YataiStub

        channel_address = channel_address.strip()
        logger.debug("Connecting YataiService gRPC server at: %s", channel_address)
        scheme, addr = parse_grpc_url(channel_address)
        header_adder_interceptor = header_client_interceptor.header_adder_interceptor(
            access_token_header, access_token
        )
        if scheme in ("grpcs", "https"):
            tls_root_ca_cert = (
                tls_root_ca_cert or certifi.where()
            )  # default: Mozilla ca cert
            with open(tls_root_ca_cert, "rb") as fb:
                ca_cert = fb.read()
            if tls_client_key:
                with open(tls_client_key, "rb") as fb:
                    tls_client_key = fb.read()
            if tls_client_cert:
                with open(tls_client_cert, "rb") as fb:
                    tls_client_cert = fb.read()
            credentials = grpc.ssl_channel_credentials(
                ca_cert, tls_client_key, tls_client_cert
            )
            channel = grpc.intercept_channel(
                grpc.secure_channel(addr, credentials), header_adder_interceptor
            )
        else:
            channel = grpc.intercept_channel(
                grpc.insecure_channel(addr), header_adder_interceptor
            )
        return YataiStub(channel)
    else:
        from bentoml.yatai.db import DB
        from bentoml.yatai.repository import create_repository
        from bentoml.yatai.yatai_service_impl import get_yatai_service_impl

        LocalYataiService = get_yatai_service_impl()

        logger.debug("Creating local YataiService instance")
        return LocalYataiService(
            repository=create_repository(
                repository_type, file_system_directory, s3_url, s3_endpoint_url, gcs_url
            ),
            database=DB(db_url),
            default_namespace=default_namespace,
        )


@inject
def start_yatai_service_grpc_server(
    db_url: str,
    grpc_port: int,
    ui_port: int,
    with_ui: bool,
    base_url: str,
    repository_type: str,
    file_system_directory: str,
    s3_url: str,
    s3_endpoint_url: str,
    gcs_url: str,
    web_ui_log_path: str = Provide[BentoMLContainer.yatai_logging_path],
):
    # Lazily import grpcio for YataiService gRPC related actions
    import grpc
    from bentoml.yatai.db import DB
    from bentoml.yatai.repository import create_repository
    from bentoml.yatai.yatai_service_impl import get_yatai_service_impl
    from bentoml.yatai.proto.yatai_service_pb2_grpc import add_YataiServicer_to_server
    from bentoml.yatai.proto.yatai_service_pb2_grpc import YataiServicer
    from bentoml.yatai.grpc_interceptor import (
        PromServerInterceptor,
        ServiceLatencyInterceptor,
    )
    from prometheus_client import start_http_server

    YataiServicerImpl = get_yatai_service_impl(YataiServicer)
    yatai_service = YataiServicerImpl(
        repository=create_repository(
            repository_type, file_system_directory, s3_url, s3_endpoint_url, gcs_url
        ),
        database=DB(db_url),
    )

    # Define interceptors here
    grpc_interceptors = [PromServerInterceptor(), ServiceLatencyInterceptor()]
    server = grpc.server(
        futures.ThreadPoolExecutor(max_workers=10), interceptors=grpc_interceptors,
    )
    add_YataiServicer_to_server(yatai_service, server)
    debug_mode = get_debug_mode()
    if debug_mode:
        try:
            logger.debug("Enabling gRPC server reflection for debugging")
            from bentoml.yatai.proto import yatai_service_pb2
            from grpc_reflection.v1alpha import reflection

            SERVICE_NAMES = (
                yatai_service_pb2.DESCRIPTOR.services_by_name["Yatai"].full_name,
                reflection.SERVICE_NAME,
            )
            reflection.enable_server_reflection(SERVICE_NAMES, server)
        except ImportError:
            logger.debug(
                "Failed to enable gRPC server reflection, missing required package: "
                '"pip install grpcio-reflection"'
            )
    server.add_insecure_port(f"[::]:{grpc_port}")

    # NOTE: the current implementation sets prometheus_port to
    # 50052 to accommodate with Makefile setups. Currently there
    # isn't a way to find the reserve_free_port dynamically inside
    # Makefile to find the free ports for prometheus_port without
    # the help of a shell scripts.
    prometheus_port = 50052
    with reserve_free_port() as port:
        prometheus_port = port
    # prevents wsgi to see prometheus_port as used
    start_http_server(prometheus_port)
    server.start()
    if with_ui:
        ensure_node_available_or_raise()
        yatai_grpc_server_address = f"localhost:{grpc_port}"
        prometheus_address = f"http://localhost:{prometheus_port}"
        async_start_yatai_service_web_ui(
            yatai_grpc_server_address,
            prometheus_address,
            ui_port,
            web_ui_log_path,
            debug_mode,
            base_url,
        )

    # We don't import _echo function from click_utils because of circular dep
    if with_ui:
        if debug_mode is True:
            ui_port = 8080
        web_ui_link = f"http://127.0.0.1:{ui_port}"
        if base_url != ".":
            web_ui_link += f"/{base_url}"
        web_ui_message = f"running on {web_ui_link}"
    else:
        web_ui_message = "off"
    if debug_mode:
        prom_ui_message = "off"
    else:
        prom_ui_message = f"running on http://127.0.0.1:{ui_port}/metrics\n"

    click.echo(
        f"* Starting BentoML YataiService gRPC Server\n"
        f'* Debug mode: { "on" if debug_mode else "off"}\n'
        f"* Web UI: {web_ui_message}\n"
        f"* Running on 127.0.0.1:{grpc_port} (Press CTRL+C to quit)\n"
        f"* Prometheus: {prom_ui_message}\n"
        f"* Help and instructions: "
        f"https://docs.bentoml.org/en/latest/guides/yatai_service.html\n"
        f'{f"* Web server log can be found here: {web_ui_log_path}" if with_ui else ""}'
        f"\n-----\n"
        f"* Usage in Python:\n"
        f'*  bento_svc.save(yatai_url="127.0.0.1:{grpc_port}")\n'
        f"*  from bentoml.yatai.client import get_yatai_client\n"
        f'*  get_yatai_client("127.0.0.1:{grpc_port}").repository.list()\n'
        f"* Usage in CLI:\n"
        f"*  bentoml list --yatai-url=127.0.0.1:{grpc_port}\n"
        f"*  bentoml containerize IrisClassifier:latest --yatai-url=127.0.0.1:"
        f"{grpc_port}\n"
        f"*  bentoml push IrisClassifier:20200918001645_CD2886 --yatai-url=127.0.0.1:"
        f"{grpc_port}\n"
        f"*  bentoml pull IrisClassifier:20200918001645_CD2886 --yatai-url=127.0.0.1:"
        f"{grpc_port}\n"
        f"*  bentoml retrieve IrisClassifier:20200918001645_CD2886 "
        f'--yatai-url=127.0.0.1:{grpc_port} --target_dir="/tmp/foo/bar"\n'
        f"*  bentoml delete IrisClassifier:20200918001645_CD2886 "
        f"--yatai-url=127.0.0.1:{grpc_port}\n"
        # TODO: simplify the example usage here once related documentation is ready
    )

    try:
        while True:
            time.sleep(_ONE_DAY_IN_SECONDS)
    except KeyboardInterrupt:
        logger.info("Terminating YataiService gRPC server..")
        server.stop(grace=None)


def _is_web_server_debug_tools_available(root_dir):
    return (
        os.path.exists(os.path.join(root_dir, "node_modules/.bin", "concurrently"))
        and os.path.exists(os.path.join(root_dir, "node_modules/.bin", "ts-node"))
        and os.path.exists(os.path.join(root_dir, "node_modules/.bin", "nodemon"))
    )


def async_start_yatai_service_web_ui(
    yatai_server_address,
    prometheus_address,
    ui_port,
    base_log_path,
    debug_mode,
    web_prefix_path,
):
    if ui_port is not None:
        ui_port = ui_port if isinstance(ui_port, str) else str(ui_port)
    web_ui_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), "web"))
    web_prefix_path = web_prefix_path.strip("/")
    if debug_mode:
        # Only when src/index.ts exists, we will run dev (nodemon)
        if os.path.exists(
            os.path.join(web_ui_dir, "src/index.ts")
        ) and _is_web_server_debug_tools_available(web_ui_dir):
            web_ui_command = [
                "npm",
                "run",
                "dev",
                "--",
                yatai_server_address,
                ui_port,
                base_log_path,
                web_prefix_path,
                prometheus_address,
            ]
        else:
            web_ui_command = [
                "node",
                "dist/bundle.js",
                yatai_server_address,
                ui_port,
                base_log_path,
                web_prefix_path,
                prometheus_address,
            ]
    else:
        if not os.path.exists(os.path.join(web_ui_dir, "dist", "bundle.js")):
            raise BentoMLException(
                "Yatai web client built is missing. "
                "Please run `npm run build` in the bentoml/yatai/web directory "
                "and then try again"
            )
        web_ui_command = [
            "node",
            "dist/bundle.js",
            yatai_server_address,
            ui_port,
            base_log_path,
            web_prefix_path,
            prometheus_address,
        ]

    web_proc = subprocess.Popen(
        web_ui_command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=web_ui_dir
    )

    is_web_proc_running = web_proc.poll() is None
    if not is_web_proc_running:
        web_proc_output = web_proc.stdout.read().decode("utf-8")
        logger.error(f"return code: {web_proc.returncode} {web_proc_output}")
        raise BentoMLException("Yatai web ui did not start properly")

    atexit.register(web_proc.terminate)


_ONE_DAY_IN_SECONDS = 60 * 60 * 24
logger = logging.getLogger(__name__)



import os

from bentoml.exceptions import MissingDependencyException
from bentoml.service.artifacts import BentoServiceArtifact
from bentoml.service.env import BentoServiceEnv
import tempfile

try:
    import paddle
    import paddle.inference as paddle_infer
except ImportError:
    paddle = None


class PaddlePaddleModelArtifact(BentoServiceArtifact):
    """
    Artifact class for saving and loading PaddlePaddle's PaddleInference model

    Args:
        name (string): name of the artifact

    Raises:
        MissingDependencyException: paddle package is required for
                                    PaddlePaddleModelArtifact

    Example usage:

    >>> import pandas as pd
    >>> import numpy as np
    >>>
    >>> from bentoml import env, artifacts, api, BentoService
    >>> from bentoml.adapters import DataframeInput
    >>> from bentoml.frameworks.paddle import PaddlePaddleModelArtifact
    >>>
    >>> @env(infer_pip_packages=True)
    >>> @artifacts([PaddlePaddleModelArtifact('model')])
    >>> class PaddleService(BentoService):
    >>>    @api(input=DataframeInput(), batch=True)
    >>>    def predict(self, df: pd.DataFrame):
    >>>        input_data = df.to_numpy().astype(np.float32)
    >>>        predictor = self.artifacts.model
    >>>
    >>>        input_names = predictor.get_input_names()
    >>>        input_handle = predictor.get_input_handle(input_names[0])
    >>>        input_handle.reshape(input_data.shape)
    >>>        input_handle.copy_from_cpu(input_data)
    >>>
    >>>        predictor.run()
    >>>
    >>>        output_names = predictor.get_output_names()
    >>>        output_handle = predictor.get_output_handle(output_names[0])
    >>>        output_data = output_handle.copy_to_cpu()
    >>>        return output_data
    >>>
    >>> service = PaddleService()
    >>>
    >>> service.pack('model', model_to_save)
    """

    def __init__(self, name: str):
        super().__init__(name)
        self._model = None
        self._predictor = None
        self._model_path = None

        if paddle is None:
            raise MissingDependencyException(
                "paddlepaddle package is required to use PaddlePaddleModelArtifact"
            )

    def pack(self, model, metadata=None):  # pylint:disable=arguments-differ
        self._model = model
        return self

    def load(self, path):
        model = paddle.jit.load(self._file_path(path))
        model = paddle.jit.to_static(model, input_spec=model._input_spec())
        return self.pack(model)

    def _file_path(self, base_path):
        return os.path.join(base_path, self.name)

    def save(self, dst):
        self._save(dst)

    def _save(self, dst):
        # Override the model path if temp dir was set
        self._model_path = self._file_path(dst)
        paddle.jit.save(self._model, self._model_path)

    def get(self):
        # Create predictor, if one doesn't exist, when inference is run
        if not self._predictor:
            # If model isn't saved, save model to a temp dir
            # because predictor init requires the path to a saved model
            if self._model_path is None:
                self._model_path = tempfile.TemporaryDirectory().name
                self._save(self._model_path)

            config = paddle_infer.Config(
                self._model_path + ".pdmodel", self._model_path + ".pdiparams"
            )
            config.enable_memory_optim()
            predictor = paddle_infer.create_predictor(config)
            self._predictor = predictor
        return self._predictor

    def set_dependencies(self, env: BentoServiceEnv):
        if env._infer_pip_packages:
            env.add_pip_packages(['paddlepaddle'])



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from abc import ABCMeta, abstractmethod


class BaseRepository(object):
    """
    BaseRepository is the interface for managing BentoML saved bundle files over either
     a file system or a cloud blob storage systems such as AWS S3 or MinIO

    A BentoML saved bundle is a standard file format that contains trained model files
    as well as serving endpoint code, input/output spec, dependency specs and deployment
    configs.
    """

    __metaclass__ = ABCMeta

    @abstractmethod
    def add(self, bento_name, bento_version):
        """
        Proposing to add a saved BentoService to target repository by providing the
        bento name and version.
        Return value is an URL(file path or s3 path for example), that is ready for
        the client to upload saved Bento files.
        """

    @abstractmethod
    def get(self, bento_name, bento_version):
        """
        Get a file path to the saved Bento files, path must be accessible form local
        machine either through NFS or pre-downloaded to local machine
        """

    @abstractmethod
    def dangerously_delete(self, bento_name, bento_version):
        """
        Deleting the Bento files that was added to this repository earlier, this may
        break existing deployments or create issues when doing deployment rollback
        """



from prometheus_client import Counter, Histogram

GRPC_SERVER_STARTED_COUNTER = Counter(
    "grpc_server_started_total",
    "Total number of RPCs started on the server.",
    ["grpc_type", "grpc_service", "grpc_method"],
)

GRPC_SERVER_STREAM_MSG_RECEIVED = Counter(
    "grpc_server_msg_received_total",
    "Total number of RPC stream messages received on the server.",
    ["grpc_type", "grpc_service", "grpc_method"],
)

GRPC_SERVER_STREAM_MSG_SENT = Counter(
    "grpc_server_msg_sent_total",
    "Total number of gRPC stream messages sent by the server.",
    ["grpc_type", "grpc_service", "grpc_method"],
)

GRPC_SERVER_HANDLED_HISTOGRAM = Histogram(
    "grpc_server_handling_seconds",
    "Histogram of response latency (seconds) of gRPC that had been application-level "
    "handled by the server.",
    ["grpc_type", "grpc_service", "grpc_method"],
)

GRPC_SERVER_HANDLED_TOTAL = Counter(
    "grpc_server_handled_total",
    "Total number of RPCs completed on the server, regardless of success or failure.",
    ["grpc_type", "grpc_service", "grpc_method", "grpc_code"],
)



import atexit
import json
import os
import platform
import shutil
import subprocess
import tempfile
import time
import zipfile
from pathlib import Path
import logging
import requests
from bentoml.exceptions import BentoMLException

logger = logging.getLogger(__name__)


def _get_command():
    """
    ngrok command based on OS
    """
    system = platform.system()
    if system == "Darwin":
        command = "ngrok"
    elif system == "Windows":
        command = "ngrok.exe"
    elif system == "Linux":
        command = "ngrok"
    else:
        raise BentoMLException("{system} is not supported".format(system=system))
    return command


def _run_ngrok(port):
    command = _get_command()
    ngrok_path = str(Path(tempfile.gettempdir(), "ngrok"))
    _download_ngrok(ngrok_path)
    executable = str(Path(ngrok_path, command))
    os.chmod(executable, 0o777)
    ngrok = subprocess.Popen([executable, 'http', str(port)])
    atexit.register(ngrok.terminate)
    localhost_url = "http://localhost:4040/api/tunnels"  # Url with tunnel details
    time.sleep(1)
    tunnel_url = requests.get(localhost_url).text  # Get the tunnel information
    j = json.loads(tunnel_url)

    tunnel_url = j['tunnels'][0]['public_url']  # Do the parsing of the get
    tunnel_url = tunnel_url.replace("https", "http")
    return tunnel_url


def _download_ngrok(ngrok_path):
    """
    Check OS and decide on ngrok download URL
    """
    if Path(ngrok_path).exists():
        return
    system = platform.system()
    if system == "Darwin":
        url = "https://bin.equinox.io/c/4VmDzA7iaHb/ngrok-stable-darwin-amd64.zip"
    elif system == "Windows":
        url = "https://bin.equinox.io/c/4VmDzA7iaHb/ngrok-stable-windows-amd64.zip"
    elif system == "Linux":
        url = "https://bin.equinox.io/c/4VmDzA7iaHb/ngrok-stable-linux-amd64.zip"
    else:
        raise Exception(f"{system} is not supported")
    download_path = _download_file(url)
    with zipfile.ZipFile(download_path, "r") as zip_ref:
        zip_ref.extractall(ngrok_path)


def _download_file(url):
    """
    Download ngrok binary file to local
    :param url:
    :return:
    """
    local_filename = url.split('/')[-1]
    r = requests.get(url, stream=True)
    download_path = str(Path(tempfile.gettempdir(), local_filename))
    with open(download_path, 'wb') as f:
        shutil.copyfileobj(r.raw, f)
    return download_path


def start_ngrok(port):
    """
    Start ngrok server
    """
    ngrok_address = _run_ngrok(port)
    logger.info(f" * Running on {ngrok_address}")
    logger.info(" * Traffic stats available on http://127.0.0.1:4040")



# Copyright 2020 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass, field
import io
import os
from typing import (
    Any,
    BinaryIO,
    Dict,
    Generic,
    Iterable,
    Iterator,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
)
import urllib
import uuid

from multidict import CIMultiDict
from werkzeug.formparser import parse_form_data
from werkzeug.http import parse_options_header

from bentoml.utils.dataclasses import json_serializer

BATCH_HEADER = "Bentoml-Is-Batch-Request"

# For non latin1 characters: https://tools.ietf.org/html/rfc8187
# Also https://github.com/benoitc/gunicorn/issues/1778
HEADER_CHARSET = 'latin1'

JSON_CHARSET = 'utf-8'


@json_serializer(fields=['uri', 'name'], compat=True)
@dataclass(frozen=False)
class FileLike:
    """
    An universal lazy-loading wrapper for file-like objects.
    It accepts URI, file path or bytes and provides interface like opened file object.

    Attributes
    ----------
    bytes : bytes, optional

    uri : str, optional
        The set of possible uris is:

        - ``file:///home/user/input.json``
        - ``http://site.com/input.csv`` (Not implemented)
        - ``https://site.com/input.csv`` (Not implemented)

    name : str, default None

    """

    bytes_: Optional[bytes] = None
    uri: Optional[str] = None
    name: Optional[str] = None

    _stream: Optional[BinaryIO] = None

    def __post_init__(self):
        if self.name is None:
            if self._stream is not None:
                self.name = getattr(self._stream, "name", None)
            elif self.uri is not None:
                p = urllib.parse.urlparse(self.uri)
                if p.scheme and p.scheme != "file":
                    raise NotImplementedError(
                        f"{self.__class__} now supports scheme 'file://' only"
                    )
                _, self.name = os.path.split(self.path)

    @property
    def path(self):
        r'''
        supports:

        /home/user/file
        C:\Python27\Scripts\pip.exe
        \\localhost\c$\WINDOWS\clock.avi
        \\networkstorage\homes\user

        https://stackoverflow.com/a/61922504/3089381
        '''
        parsed = urllib.parse.urlparse(self.uri)
        raw_path = urllib.request.url2pathname(urllib.parse.unquote(parsed.path))
        host = "{0}{0}{mnt}{0}".format(os.path.sep, mnt=parsed.netloc)
        path = os.path.abspath(os.path.join(host, raw_path))
        return path

    @property
    def stream(self):
        if self._stream is not None:
            pass
        elif self.bytes_ is not None:
            self._stream = io.BytesIO(self.bytes_)
        elif self.uri is not None:
            self._stream = open(self.path, "rb")
        else:
            return io.BytesIO()
        return self._stream

    def read(self, size=-1):
        # TODO: also write to log
        return self.stream.read(size)

    def seek(self, pos):
        return self.stream.seek(pos)

    def tell(self):
        return self.stream.tell()

    def close(self):
        if self._stream is not None:
            self._stream.close()

    def __del__(self):
        if getattr(self, "_stream", None) and not self._stream.closed:
            self._stream.close()


class HTTPHeaders(CIMultiDict):
    """
    A case insensitive mapping of HTTP headers' keys and values.
    It also parses several commonly used fields for easier access.

    Attributes
    ----------
    content_type : str
        The value of ``Content-Type``, for example:
        - ``application/json``
        - ``text/plain``
        - ``text/csv``

    charset : str
        The charset option of ``Content-Type``

    content_encoding : str
        The charset option of ``Content-Encoding``

    Methods
    -------
    from_dict : create a HTTPHeaders object from a dict

    from_sequence : create a HTTPHeaders object from a list/tuple

    """

    @property
    def content_type(self) -> str:
        return parse_options_header(self.get('content-type'))[0].lower()

    @property
    def charset(self) -> Optional[str]:
        _, options = parse_options_header(self.get('content-type'))
        charset = options.get('charset', None)
        assert charset is None or isinstance(charset, str)
        return charset

    @property
    def content_encoding(self) -> str:
        return parse_options_header(self.get('content-encoding'))[0].lower()

    @property
    def is_batch_input(self) -> Optional[bool]:
        hv = parse_options_header(self.get(BATCH_HEADER))[0].lower()
        return hv == "true" if hv else None

    @classmethod
    def from_dict(cls, d: Mapping[str, str]):
        return cls(d)

    @classmethod
    def from_sequence(cls, seq: Sequence[Tuple[str, str]]):
        return cls(seq)

    def to_json(self):
        return tuple(self.items())


@dataclass
class HTTPRequest:
    """
    A common HTTP Request object.
    It also parses several commonly used fields for easier access.

    Attributes
    ----------
    headers : HTTPHeaders

    body : bytes

    """

    headers: HTTPHeaders = HTTPHeaders()
    body: bytes = b""

    def __post_init__(self):
        if self.headers is None:
            self.headers = HTTPHeaders()
        elif isinstance(self.headers, dict):
            self.headers = HTTPHeaders.from_dict(self.headers)
        elif isinstance(self.headers, (tuple, list)):
            self.headers = HTTPHeaders.from_sequence(self.headers)

    @classmethod
    def parse_form_data(cls, self):
        if not self.body:
            return None, None, {}
        environ = {
            'wsgi.input': io.BytesIO(self.body),
            'CONTENT_LENGTH': len(self.body),
            'CONTENT_TYPE': self.headers.get('content-type', ''),
            'REQUEST_METHOD': 'POST',
        }
        stream, form, files = parse_form_data(environ, silent=False)
        wrapped_files = {
            k: FileLike(_stream=f, name=f.filename) for k, f in files.items()
        }
        return stream, form, wrapped_files

    @classmethod
    def from_flask_request(cls, request):
        return cls(
            tuple((k, v) for k, v in request.headers.items()), request.get_data(),
        )

    def to_flask_request(self):
        from werkzeug.wrappers import Request

        return Request.from_values(
            input_stream=io.BytesIO(self.body),
            content_length=len(self.body),
            headers=self.headers,
        )


@dataclass
class HTTPResponse:
    status: int = 200
    headers: HTTPHeaders = HTTPHeaders()
    body: Optional[bytes] = b""

    @classmethod
    def new(
        cls,
        status: int = 200,
        headers: Union[HTTPHeaders, dict, tuple, list] = None,
        body: bytes = b"",
    ):
        if headers is None:
            headers = HTTPHeaders()
        elif isinstance(headers, dict):
            headers = HTTPHeaders.from_dict(headers)
        elif isinstance(headers, (tuple, list)):
            headers = HTTPHeaders.from_sequence(headers)
        return cls(status, headers, body)

    def __post_init__(self):
        if self.headers is None:
            self.headers = HTTPHeaders()
        elif isinstance(self.headers, dict):
            self.headers = HTTPHeaders.from_dict(self.headers)
        elif isinstance(self.headers, (tuple, list)):
            self.headers = HTTPHeaders.from_sequence(self.headers)

    def to_flask_response(self):
        import flask

        return flask.Response(
            status=self.status, headers=tuple(self.headers.items()), response=self.body
        )


# https://tools.ietf.org/html/rfc7159#section-3
JsonSerializable = Union[bool, None, Dict, List, int, float, str]

# https://docs.aws.amazon.com/lambda/latest/dg/python-handler.html
AwsLambdaEvent = Union[Dict, List, str, int, float, None]

Input = TypeVar("Input")
Output = TypeVar("Output")

ApiFuncArgs = TypeVar("ApiFuncArgs")
BatchApiFuncArgs = TypeVar("BatchApiFuncArgs")
ApiFuncReturnValue = TypeVar("ApiFuncReturnValue")
BatchApiFuncReturnValue = TypeVar("BatchApiFuncReturnValue")


@json_serializer(compat=True)
@dataclass
class InferenceResult(Generic[Output]):
    """
    The data structure that returned by BentoML API server.
    Contains result data and context like HTTP headers.
    """

    version: int = 0

    # payload
    data: Optional[Output] = None
    err_msg: str = ''

    # meta
    task_id: Optional[str] = None

    # context
    http_status: int = 501
    http_headers: HTTPHeaders = HTTPHeaders()
    aws_lambda_event: Optional[dict] = None
    cli_status: Optional[int] = 0

    def __post_init__(self):
        if self.http_headers is None:
            self.http_headers = HTTPHeaders()
        elif isinstance(self.http_headers, dict):
            self.http_headers = HTTPHeaders.from_dict(self.http_headers)
        elif isinstance(self.http_headers, (tuple, list)):
            self.http_headers = HTTPHeaders.from_sequence(self.http_headers)

    @classmethod
    def complete_discarded(
        cls, tasks: Iterable['InferenceTask'], results: Iterable['InferenceResult'],
    ) -> Iterator['InferenceResult']:
        """
        Generate InferenceResults based on successful inference results and
        fallback results of discarded tasks.

        """
        iterable_results = iter(results)
        try:
            for task in tasks:
                if task.is_discarded:
                    assert task.error
                    yield task.error
                else:
                    yield next(iterable_results)
        except StopIteration:
            raise StopIteration(
                'The results does not match the number of tasks'
            ) from None


@json_serializer(compat=True)
@dataclass
class InferenceError(InferenceResult):
    """
    The default InferenceResult when errors happened.
    """

    # context
    http_status: int = 500
    cli_status: int = 1


@json_serializer(compat=True)
@dataclass
class InferenceTask(Generic[Input]):
    """
    The data structure passed to the BentoML API server for inferring.
    Contains payload data and context like HTTP headers or CLI args.
    """

    version: int = 0

    # payload
    data: Optional[Input] = None
    error: Optional[InferenceResult] = None

    # meta
    task_id: str = field(default_factory=lambda: str(uuid.uuid4()))
    is_discarded: bool = False
    batch: Optional[int] = None

    # context
    http_method: Optional[str] = None
    http_headers: HTTPHeaders = HTTPHeaders()
    aws_lambda_event: Optional[dict] = None
    cli_args: Optional[Sequence[str]] = None
    inference_job_args: Optional[Mapping[str, Any]] = None

    def discard(self, err_msg="", **context):
        """
        Discard this task. All subsequent steps will be skipped.

        Parameters
        ----------
        err_msg: str
            The reason why this task got discarded. It would be the body of
            HTTP Response, a field in AWS lambda event or CLI stderr message.

        *other contexts
            Other contexts of the fallback ``InferenceResult``
        """
        self.is_discarded = True
        self.error = InferenceError(err_msg=err_msg, **context)
        return self



"""add labels table

Revision ID: 719dd2aacc9a
Revises: a6b00ae45279
Create Date: 2020-08-05 22:41:30.611193

"""
from alembic import op
import sqlalchemy as sa

from bentoml.yatai.db.stores.label import Label

# revision identifiers, used by Alembic.
revision = '719dd2aacc9a'
down_revision = 'a6b00ae45279'
branch_labels = None
depends_on = None


def upgrade():
    bind = op.get_bind()
    Label.__table__.create(bind)
    with op.batch_alter_table('deployments') as batch_op:
        batch_op.drop_column('labels')


def downgrade():
    op.add_column('deployments', sa.Column('labels', sa.JSON, default={}))
    op.drop_table('labels')



"""Interceptor a client call with prometheus"""

from timeit import default_timer

import grpc
from bentoml.yatai.metrics import (
    GRPC_SERVER_HANDLED_HISTOGRAM,
    GRPC_SERVER_HANDLED_TOTAL,
    GRPC_SERVER_STARTED_COUNTER,
    GRPC_SERVER_STREAM_MSG_RECEIVED,
    GRPC_SERVER_STREAM_MSG_SENT,
)
from bentoml.yatai.utils import (
    get_method_type,
    parse_method_name,
    wrap_interator_inc_counter,
)
from bentoml.utils.usage_stats import track


YATAI_GRPC_USAGE_EVENT_NAME = "yatai-grpc-call"


def _wrap_rpc_behaviour(handler, fn):
    """Returns a new rpc handler that wraps the given function"""
    if handler is None:
        return None

    if handler.request_streaming and handler.response_streaming:
        behavior_fn = handler.stream_stream
        handler_factory = grpc.stream_stream_rpc_method_handler
    elif handler.request_streaming and not handler.response_streaming:
        behavior_fn = handler.stream_unary
        handler_factory = grpc.stream_unary_rpc_method_handler
    elif not handler.request_streaming and handler.response_streaming:
        behavior_fn = handler.unary_stream
        handler_factory = grpc.unary_stream_rpc_method_handler
    else:
        behavior_fn = handler.unary_unary
        handler_factory = grpc.unary_unary_rpc_method_handler

    return handler_factory(
        fn(behavior_fn, handler.request_streaming, handler.response_streaming),
        request_deserializer=handler.request_deserializer,
        response_serializer=handler.response_serializer,
    )


class PromServerInterceptor(grpc.ServerInterceptor):  # pylint: disable=W0232
    """
    Interceptor for handling client call with metrics to prometheus
    This implementation is referred to:
    https://grpc.github.io/grpc/python/grpc.html#service-side-interceptor
    https://github.com/lchenn/py-grpc-prometheus

    Server metrics are exposed by adding the interceptor when gRPC server is started.
    Refers to
    tests/yatai/grpc_testing_service.py for a mock gRPC service
    tests/yatai/client/test_grpc_server_interceptor.py for an implementation's example

    ```python

        import grpc
        from prometheus_client import start_http_server
        from bentoml.yatai.client.interceptor import PromServerInterceptor
        ...
        server = grpc.server(
            futures.ThreadPoolExecutor(max_workers=1),
            interceptors=(PromServerInterceptor(), )
            )
        start_http_server(self.prom_port)
    ```
    """

    def intercept_service(self, continuation, handler_call_details):

        # handles if call_details are None
        handler = continuation(handler_call_details)
        if handler is None:
            return handler

        # unary
        if handler.request_streaming or handler.response_streaming:
            return handler

        mn, ok = parse_method_name(handler_call_details.method)
        grpc_service_name = mn.fully_qualified_service
        grpc_method_name = mn.method

        if not ok:
            return continuation(handler_call_details)

        def metrics_wrapper(behaviour, request_streaming, response_streaming):
            def new_behaviour(request_or_iterator, servicer_context):
                grpc_type = get_method_type(request_streaming, response_streaming)

                try:
                    if request_streaming:
                        request_or_iterator = wrap_interator_inc_counter(
                            request_or_iterator,
                            GRPC_SERVER_STREAM_MSG_RECEIVED,
                            grpc_type,
                            grpc_service_name,
                            grpc_method_name,
                        )
                        request_or_iterator = wrap_interator_inc_counter(
                            request_or_iterator,
                            GRPC_SERVER_STREAM_MSG_SENT,
                            grpc_type,
                            grpc_service_name,
                            grpc_method_name,
                        )
                    else:
                        GRPC_SERVER_STARTED_COUNTER.labels(
                            grpc_type=grpc_type,
                            grpc_service=grpc_service_name,
                            grpc_method=grpc_method_name,
                        ).inc()

                    # Invoke the original RPC behaviour
                    response_or_iterator = behaviour(
                        request_or_iterator, servicer_context
                    )

                    if response_streaming:
                        sent_metrics = GRPC_SERVER_STREAM_MSG_SENT
                        response_or_iterator = wrap_interator_inc_counter(
                            response_or_iterator,
                            GRPC_SERVER_STREAM_MSG_RECEIVED,
                            grpc_type,
                            grpc_service_name,
                            grpc_method_name,
                        )
                        response_or_iterator = wrap_interator_inc_counter(
                            response_or_iterator,
                            sent_metrics,
                            grpc_type,
                            grpc_service_name,
                            grpc_method_name,
                        )
                    else:
                        GRPC_SERVER_HANDLED_TOTAL.labels(
                            grpc_type=grpc_type,
                            grpc_service=grpc_service_name,
                            grpc_method=grpc_method_name,
                            grpc_code=self._compute_status_code(servicer_context).name,
                        ).inc()

                    return response_or_iterator

                except grpc.RpcError as e:
                    GRPC_SERVER_HANDLED_TOTAL.labels(
                        grpc_type=grpc_type,
                        grpc_service=grpc_service_name,
                        grpc_method=grpc_method_name,
                        grpc_code=self._compute_error_code(e),
                    ).inc()
                    raise e

            return new_behaviour

        optional_any = _wrap_rpc_behaviour(
            continuation(handler_call_details), metrics_wrapper
        )
        return optional_any

    def _compute_status_code(self, servicer_context: grpc.ServicerContext):
        if servicer_context._state.client == "cancelled":
            return grpc.StatusCode.CANCELLED

        if servicer_context._state.code is None:
            return grpc.StatusCode.OK

        return servicer_context._state.code

    def _compute_error_code(self, grpc_exception):
        if isinstance(grpc_exception, grpc.Call):
            return grpc_exception.code().name

        return grpc.StatusCode.UNKNOWN.name


class ServiceLatencyInterceptor(grpc.ServerInterceptor):  # pylint: disable=W0232
    """Interceptor to handle service latency calls"""

    def intercept_service(self, continuation, handler_call_details):

        handler = continuation(handler_call_details)
        if handler is None:
            return handler

        # unary
        if handler.request_streaming or handler.response_streaming:
            return handler

        mn, ok = parse_method_name(handler_call_details.method)
        grpc_service_name = mn.fully_qualified_service
        grpc_method_name = mn.method

        if not ok:
            return continuation(handler_call_details)

        # pylint: disable=W0613
        def latency_wrapper(behaviour, request_streaming, response_streaming):
            def new_behaviour(request_or_iterator, servicer_context):
                start = default_timer()
                try:
                    return behaviour(request_or_iterator, servicer_context)
                finally:
                    duration = max(default_timer() - start, 0)
                    GRPC_SERVER_HANDLED_HISTOGRAM.labels(
                        grpc_type='UNARY',
                        grpc_service=grpc_service_name,
                        grpc_method=grpc_method_name,
                    ).observe(duration)
                    track(
                        YATAI_GRPC_USAGE_EVENT_NAME,
                        {"method": grpc_method_name, "duration": duration},
                    )

            return new_behaviour

        return _wrap_rpc_behaviour(continuation(handler_call_details), latency_wrapper)



import os

from bentoml.exceptions import InvalidArgument, MissingDependencyException
from bentoml.service.artifacts import BentoServiceArtifact
from bentoml.service.env import BentoServiceEnv

LIGHTGBM_MODEL_EXTENTION = ".txt"


class LightGBMModelArtifact(BentoServiceArtifact):
    """
    Artifact class for saving and loading LightGBM Model

    Args:
        name (string): name of the artifact

    Raises:
        MissingDependencyException: lightgbm package is required for using
            LightGBMModelArtifact
        InvalidArgument: invalid argument type, model being packed must be instance of
            lightgbm.Booster

    Example usage:

    >>> import lightgbm as lgb
    >>> # Prepare data
    >>> train_data = lgb.Dataset(...)
    >>> # train model
    >>> model = lgb.train(train_set=train_data, ...)
    >>>
    >>> import bentoml
    >>> from bentoml.frameworks.lightgbm import LightGBMModelArtifact
    >>> from bentoml.adapters import DataframeInput
    >>>
    >>> @bentoml.artifacts([LightGBMModelArtifact('model')])
    >>> @bentoml.env(infer_pip_packages=True)
    >>> class LgbModelService(bentoml.BentoService):
    >>>
    >>>     @bentoml.api(input=DataframeInput(), batch=True)
    >>>     def predict(self, df):
    >>>         return self.artifacts.model.predict(df)
    >>>
    >>> svc = LgbModelService()
    >>> svc.pack('model', model)
    """

    def __init__(self, name):
        super(LightGBMModelArtifact, self).__init__(name)
        self._model = None

    def _model_file_path(self, base_path):
        return os.path.join(base_path, self.name + LIGHTGBM_MODEL_EXTENTION)

    def pack(self, model, metadata=None):  # pylint:disable=arguments-differ
        try:
            import lightgbm as lgb
        except ImportError:
            raise MissingDependencyException(
                "lightgbm package is required to use LightGBMModelArtifact"
            )

        if not isinstance(model, lgb.Booster):
            raise InvalidArgument(
                "Expect `model` argument to be a `lightgbm.Booster` instance"
            )

        self._model = model
        return self

    def load(self, path):
        try:
            import lightgbm as lgb
        except ImportError:
            raise MissingDependencyException(
                "lightgbm package is required to use LightGBMModelArtifact"
            )
        bst = lgb.Booster(model_file=self._model_file_path(path))

        return self.pack(bst)

    def set_dependencies(self, env: BentoServiceEnv):
        if env._infer_pip_packages:
            env.add_pip_packages(['lightgbm'])

    def save(self, dst):
        return self._model.save_model(self._model_file_path(dst))

    def get(self):
        return self._model



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# List of APIs for accessing remote or local yatai service via Python


import logging
from typing import TYPE_CHECKING, Optional

from bentoml.utils import cached_property
from bentoml.yatai.client.bento_repository_api import BentoRepositoryAPIClient
from bentoml.yatai.client.deployment_api import DeploymentAPIClient
from bentoml.yatai.yatai_service import get_yatai_service

if TYPE_CHECKING:
    from bentoml.yatai.proto.yatai_service_pb2_grpc import YataiStub

logger = logging.getLogger(__name__)


class YataiClient:
    """Python Client for interacting with YataiService"""

    def __init__(self, yatai_service: Optional["YataiStub"] = None):
        self.yatai_service = yatai_service if yatai_service else get_yatai_service()
        self.bento_repository_api_client = None
        self.deployment_api_client = None

    @cached_property
    def repository(self) -> "BentoRepositoryAPIClient":
        return BentoRepositoryAPIClient(self.yatai_service)

    @cached_property
    def deployment(self) -> "DeploymentAPIClient":
        return DeploymentAPIClient(self.yatai_service)


def get_yatai_client(yatai_url: str = None) -> "YataiClient":
    """
    Args:
        yatai_url (`str`):
            Yatai Service URL address.

    Returns:
        :obj:`~YataiClient`, a python client to interact with :obj:`Yatai` gRPC server.

    Example::

        from bentoml.yatai.client import get_yatai_client

        custom_url = 'https://remote.yatai:50050'
        yatai_client = get_yatai_client(custom_url)
    """  # noqa: E501

    yatai_service = get_yatai_service(channel_address=yatai_url)
    return YataiClient(yatai_service=yatai_service)



from bentoml.yatai.grpc_interceptor.prom_server_interceptor import (
    PromServerInterceptor,
    ServiceLatencyInterceptor,
)

__all__ = [
    "ServiceLatencyInterceptor",
    "PromServerInterceptor",
]



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from urllib.parse import urlparse

logger = logging.getLogger(__name__)


def is_gcs_url(url):
    """
    Check if the url is a gcs url
    'gs://' is the standard way for Google Cloud URI
    Source: https://cloud.google.com/storage/docs/gsutil
    """
    try:
        return urlparse(url).scheme in ["gs"]
    except ValueError:
        return False



import enum
import re

from sqlalchemy import UniqueConstraint, Column, Integer, String, and_, Enum

from bentoml.exceptions import YataiLabelException, InvalidArgument
from bentoml.yatai.db import Base
from bentoml.yatai.proto.label_selectors_pb2 import LabelSelectors


class RESOURCE_TYPE(enum.Enum):
    deployment = 1
    bento = 2


def _validate_labels(labels):
    """
    Validate labels key value format is:
        * Between 3 and 63 characters
        * Consist of alphanumeric, dash (-), period (.), and underscore (_)
        * Start and end with alphanumeric
    Args:
        labels: Dictionary

    Returns:
    Raise:
        InvalidArgument
    """
    if not isinstance(labels, dict):
        raise InvalidArgument('BentoService labels must be a dictionary')

    pattern = re.compile("^(([A-Za-z0-9][-A-Za-z0-9_.]*)?[A-Za-z0-9])?$")
    for key in labels:
        if (
            not (63 >= len(key) >= 3)
            or not (63 >= len(labels[key]) >= 3)
            or not pattern.match(key)
            or not pattern.match(labels[key])
        ):
            raise InvalidArgument(
                f'Invalide label {key}:{labels[key]}. Valid label key and value must '
                f'be between 3 to 63 characters and must be begin and end with '
                f'an alphanumeric character ([a-z0-9A-Z]) with dashes (-), '
                f'underscores (_), and dots (.).'
            )


class Label(Base):
    __tablename__ = 'labels'
    __table_args__ = tuple(
        UniqueConstraint(
            'resource_type',
            'resource_id',
            'key',
            name='_resource_type_resource_id_key_uc',
        )
    )
    id = Column(Integer, primary_key=True)
    resource_type = Column(Enum(RESOURCE_TYPE))
    resource_id = Column(Integer, nullable=False)
    key = Column(String, nullable=False)
    value = Column(String, nullable=False)


class LabelStore(object):
    @staticmethod
    def add(sess, resource_type, resource_id, labels):
        label_objects = []
        for key in labels:
            label_obj = Label()
            label_obj.resource_type = resource_type
            label_obj.resource_id = resource_id
            label_obj.key = key
            label_obj.value = labels[key]
            label_objects.append(label_obj)
        return sess.add_all(label_objects)

    @staticmethod
    def add_or_update(sess, resource_type, resource_id, labels):
        label_rows = (
            sess.query(Label)
            .filter(
                Label.resource_type == resource_type, Label.resource_id == resource_id
            )
            .all()
        )
        if len(label_rows) == 0:
            return LabelStore.add(sess, resource_type, resource_id, labels)
        else:
            for row in label_rows:
                if labels.get(row.key, None) is not None:
                    row.value = labels[row.key]

    @staticmethod
    def get(sess, resource_type, resource_id):
        labels = (
            sess.query(Label)
            .filter_by(resource_type=resource_type, resource_id=resource_id)
            .all()
        )
        return {row.key: row.value for row in labels}

    @staticmethod
    def list(sess, resource_type, resource_ids):
        label_rows = (
            sess.query(Label)
            .filter(
                and_(
                    Label.resource_type == resource_type,
                    Label.resource_id.in_(resource_ids),
                )
            )
            .all()
        )
        labels = {}
        for label in label_rows:
            if not labels.get(str(label.resource_id)):
                labels[str(label.resource_id)] = {}
            labels[str(label.resource_id)][label.key] = label.value
        return labels

    @staticmethod
    def delete(sess, resource_type, resource_id, labels=None):
        filters = [
            Label.resource_id == resource_id,
            Label.resource_type == resource_type,
        ]
        if labels is not None:
            filters.append(Label.key.in_(list(labels.keys())))
        return sess.query(Label).filter(and_(*filters)).delete()

    @staticmethod
    def filter_query(sess, resource_type, label_selectors):
        query = sess.query(Label.resource_id)
        query = query.filter_by(resource_type=resource_type)
        filters = []
        for key in label_selectors.match_labels:
            filters.append(
                sess.query(Label.resource_id).filter(
                    and_(
                        Label.key == key,
                        Label.value == label_selectors.match_labels[key],
                    )
                )
            )
        for expression in label_selectors.match_expressions:
            if (
                expression.operator
                == LabelSelectors.LabelSelectorExpression.OPERATOR_TYPE.In
            ):
                filters.append(
                    sess.query(Label.resource_id).filter(
                        and_(
                            Label.key == expression.key,
                            Label.value.in_(expression.values),
                        )
                    )
                )
            elif (
                expression.operator
                == LabelSelectors.LabelSelectorExpression.OPERATOR_TYPE.NotIn
            ):
                filters.append(
                    sess.query(Label.resource_id).filter(
                        and_(
                            Label.key == expression.key,
                            ~Label.value.in_(expression.values),
                        )
                    )
                )
            elif (
                expression.operator
                == LabelSelectors.LabelSelectorExpression.OPERATOR_TYPE.Exists
            ):
                filters.append(
                    sess.query(Label.resource_id).filter(Label.key == expression.key)
                )
            elif (
                expression.operator
                == LabelSelectors.LabelSelectorExpression.OPERATOR_TYPE.DoesNotExist
            ):
                filters.append(
                    sess.query(Label.resource_id).filter(Label.key != expression.key)
                )
            else:
                raise YataiLabelException(
                    f'Unrecognized operator: "{expression.operator}"'
                )
        query = query.intersect(*filters)
        result = query.all()
        return [row[0] for row in result]



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import datetime
from contextlib import contextmanager

from sqlalchemy import (
    Column,
    String,
    Integer,
    DateTime,
    JSON,
    UniqueConstraint,
    desc,
)
from sqlalchemy.orm.exc import NoResultFound
from google.protobuf.json_format import ParseDict

from bentoml.exceptions import YataiDeploymentException
from bentoml.yatai.deployment import ALL_NAMESPACE_TAG
from bentoml.yatai.db.base import Base
from bentoml.yatai.db.stores.label import (
    LabelStore,
    RESOURCE_TYPE,
)
from bentoml.yatai.proto import deployment_pb2
from bentoml.yatai.proto.deployment_pb2 import DeploymentSpec, ListDeploymentsRequest
from bentoml.utils import ProtoMessageToDict

logger = logging.getLogger(__name__)


class Deployment(Base):
    __tablename__ = 'deployments'
    __table_args__ = tuple(
        UniqueConstraint('name', 'namespace', name='_name_namespace_uc')
    )

    id = Column(Integer, primary_key=True)
    name = Column(String, nullable=False)
    namespace = Column(String, nullable=False)

    spec = Column(JSON, nullable=False, default={})
    state = Column(JSON, nullable=False, default={})
    annotations = Column(JSON, nullable=False, default={})

    created_at = Column(DateTime, default=datetime.datetime.utcnow)
    last_updated_at = Column(DateTime, default=datetime.datetime.utcnow)


def _deployment_pb_to_orm_obj(deployment_pb, deployment_obj=Deployment()):
    deployment_obj.name = deployment_pb.name
    deployment_obj.namespace = deployment_pb.namespace
    deployment_obj.spec = ProtoMessageToDict(deployment_pb.spec)
    deployment_obj.state = ProtoMessageToDict(deployment_pb.state)
    deployment_obj.annotations = dict(deployment_pb.annotations)
    deployment_obj.created_at = deployment_pb.created_at.ToDatetime()
    deployment_obj.last_updated_at = deployment_pb.last_updated_at.ToDatetime()
    return deployment_obj


def _deployment_orm_obj_to_pb(deployment_obj, labels=None):
    deployment_pb = deployment_pb2.Deployment(
        name=deployment_obj.name,
        namespace=deployment_obj.namespace,
        spec=ParseDict(deployment_obj.spec, deployment_pb2.DeploymentSpec()),
        state=ParseDict(deployment_obj.state, deployment_pb2.DeploymentState()),
        annotations=deployment_obj.annotations,
    )
    deployment_pb.created_at.FromDatetime(deployment_obj.created_at)
    if deployment_obj.last_updated_at:
        deployment_pb.last_updated_at.FromDatetime(deployment_obj.last_updated_at)
    if labels is not None:
        deployment_pb.labels.update(labels)
    return deployment_pb


class DeploymentStore(object):
    @staticmethod
    def insert_or_update(sess, deployment_pb):
        try:
            deployment_obj = (
                sess.query(Deployment)
                .filter_by(name=deployment_pb.name, namespace=deployment_pb.namespace)
                .one()
            )
            if deployment_obj:
                # updating deployment record in db
                _deployment_pb_to_orm_obj(deployment_pb, deployment_obj)
                if deployment_pb.labels:
                    LabelStore.add_or_update(
                        sess,
                        RESOURCE_TYPE.deployment,
                        deployment_obj.id,
                        deployment_pb.labels,
                    )
        except NoResultFound:
            deployment_orm_obj = _deployment_pb_to_orm_obj(deployment_pb)
            sess.add(deployment_orm_obj)
            if deployment_pb.labels:
                deployment_row = (
                    sess.query(Deployment)
                    .filter_by(
                        name=deployment_orm_obj.name,
                        namespace=deployment_orm_obj.namespace,
                    )
                    .one()
                )
                LabelStore.add(
                    sess,
                    RESOURCE_TYPE.deployment,
                    deployment_row.id,
                    deployment_pb.labels,
                )

    @staticmethod
    @contextmanager
    def update_deployment(sess, name, namespace):
        try:
            deployment_obj = (
                sess.query(Deployment).filter_by(name=name, namespace=namespace).one()
            )
            yield deployment_obj
        except NoResultFound:
            yield None

    @staticmethod
    def get(sess, name, namespace):
        try:
            deployment_obj = (
                sess.query(Deployment).filter_by(name=name, namespace=namespace).one()
            )
            labels = LabelStore.get(sess, RESOURCE_TYPE.deployment, deployment_obj.id)
        except NoResultFound:
            return None

        return _deployment_orm_obj_to_pb(deployment_obj, labels)

    @staticmethod
    def delete(sess, name, namespace):
        try:
            deployment = (
                sess.query(Deployment).filter_by(name=name, namespace=namespace).one()
            )
            LabelStore.delete(
                sess, resource_type=RESOURCE_TYPE.deployment, resource_id=deployment.id,
            )
            return sess.delete(deployment)
        except NoResultFound:
            raise YataiDeploymentException(
                "Deployment '%s' in namespace: '%s' is not found" % name, namespace
            )

    @staticmethod
    def list(
        sess,
        namespace,
        operator=None,
        label_selectors=None,
        offset=None,
        limit=None,
        order_by=ListDeploymentsRequest.created_at,
        ascending_order=False,
    ):
        query = sess.query(Deployment)
        order_by = ListDeploymentsRequest.SORTABLE_COLUMN.Name(order_by)
        order_by_field = getattr(Deployment, order_by)
        order_by_action = order_by_field if ascending_order else desc(order_by_field)
        query = query.order_by(order_by_action)
        if label_selectors.match_labels or label_selectors.match_expressions:
            deployment_ids = LabelStore.filter_query(
                sess, RESOURCE_TYPE.deployment, label_selectors
            )
            query.filter(Deployment.id.in_(deployment_ids))
        if namespace != ALL_NAMESPACE_TAG:  # else query all namespaces
            query = query.filter_by(namespace=namespace)
        if operator:
            operator_name = DeploymentSpec.DeploymentOperator.Name(operator)
            query = query.filter(
                Deployment.spec['operator'].as_string().contains(operator_name)
            )
        # We are not defaulting limit to 200 in the signature,
        # because protobuf will pass 0 as value
        limit = limit or 200
        # Limit and offset need to be called after order_by filter/filter_by is
        # called
        query = query.limit(limit)
        if offset:
            query = query.offset(offset)
        query_result = query.all()
        deployment_ids = [deployment_obj.id for deployment_obj in query_result]
        labels = LabelStore.list(sess, RESOURCE_TYPE.deployment, deployment_ids)

        return [
            _deployment_orm_obj_to_pb(
                deployment_obj, labels.get(str(deployment_obj.id))
            )
            for deployment_obj in query_result
        ]



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# List of APIs for accessing remote or local yatai service via Python


import logging
import time

from bentoml.utils import status_pb_to_error_code_and_message
from bentoml.yatai.client.label_utils import generate_gprc_labels_selector
from bentoml.yatai.deployment import ALL_NAMESPACE_TAG
from bentoml.yatai.proto.deployment_pb2 import (
    ApplyDeploymentRequest,
    DescribeDeploymentRequest,
    GetDeploymentRequest,
    DeploymentSpec,
    DeleteDeploymentRequest,
    ListDeploymentsRequest,
    Deployment,
    DeploymentState,
)
from bentoml.exceptions import BentoMLException, YataiDeploymentException
from bentoml.yatai.proto import status_pb2
from bentoml.yatai.deployment_utils import (
    deployment_yaml_string_to_pb,
    deployment_dict_to_pb,
)

logger = logging.getLogger(__name__)

WAIT_TIMEOUT_LIMIT = 600
WAIT_TIME = 5


class DeploymentAPIClient:
    def __init__(self, yatai_service):
        self.yatai_service = yatai_service

    def list(
        self,
        limit=None,
        offset=None,
        labels=None,
        namespace=None,
        is_all_namespaces=False,
        operator=None,
        order_by=None,
        ascending_order=None,
    ):
        if is_all_namespaces:
            if namespace is not None:
                logger.warning(
                    'Ignoring `namespace=%s` due to the --all-namespace flag presented',
                    namespace,
                )
            namespace = ALL_NAMESPACE_TAG
        if isinstance(operator, str):
            if operator == 'sagemaker':
                operator = DeploymentSpec.AWS_SAGEMAKER
            elif operator == 'lambda':
                operator = DeploymentSpec.AWS_LAMBDA
            elif operator == DeploymentSpec.AZURE_FUNCTIONS:
                operator = 'azure-functions'
            elif operator == "ec2":
                operator = DeploymentSpec.AWS_EC2
            else:
                raise BentoMLException(f'Unrecognized operator {operator}')

        list_deployment_request = ListDeploymentsRequest(
            limit=limit,
            offset=offset,
            namespace=namespace,
            operator=operator,
            order_by=order_by,
            ascending_order=ascending_order,
        )
        if labels is not None:
            generate_gprc_labels_selector(
                list_deployment_request.label_selectors, labels
            )
        return self.yatai_service.ListDeployments(list_deployment_request)

    def get(self, namespace, name):
        return self.yatai_service.GetDeployment(
            GetDeploymentRequest(deployment_name=name, namespace=namespace)
        )

    def describe(self, namespace, name):
        return self.yatai_service.DescribeDeployment(
            DescribeDeploymentRequest(deployment_name=name, namespace=namespace)
        )

    def delete(self, deployment_name, namespace, force_delete=False):
        return self.yatai_service.DeleteDeployment(
            DeleteDeploymentRequest(
                deployment_name=deployment_name,
                namespace=namespace,
                force_delete=force_delete,
            )
        )

    def create(self, deployment_info, wait):
        from bentoml.yatai.validator import validate_deployment_pb

        if isinstance(deployment_info, dict):
            deployment_pb = deployment_dict_to_pb(deployment_info)
        elif isinstance(deployment_info, str):
            deployment_pb = deployment_yaml_string_to_pb(deployment_info)
        elif isinstance(deployment_info, Deployment):
            deployment_pb = deployment_info
        else:
            raise YataiDeploymentException(
                'Unexpected argument type, expect deployment info to be str in yaml '
                'format or a dict or a deployment protobuf obj, instead got: {}'.format(
                    str(type(deployment_info))
                )
            )

        validation_errors = validate_deployment_pb(deployment_pb)
        if validation_errors:
            raise YataiDeploymentException(
                f'Failed to validate deployment {deployment_pb.name}: '
                f'{validation_errors}'
            )
        # Make sure there is no active deployment with the same deployment name
        get_deployment_pb = self.yatai_service.GetDeployment(
            GetDeploymentRequest(
                deployment_name=deployment_pb.name, namespace=deployment_pb.namespace
            )
        )
        if get_deployment_pb.status.status_code != status_pb2.Status.NOT_FOUND:
            raise BentoMLException(
                f'Deployment "{deployment_pb.name}" already existed, use Update or '
                f'Apply for updating existing deployment, delete the deployment, '
                f'or use a different deployment name'
            )
        apply_result = self.yatai_service.ApplyDeployment(
            ApplyDeploymentRequest(deployment=deployment_pb)
        )
        if apply_result.status.status_code != status_pb2.Status.OK:
            error_code, error_message = status_pb_to_error_code_and_message(
                apply_result.status
            )
            raise YataiDeploymentException(f'{error_code}:{error_message}')
        if wait:
            self._wait_deployment_action_complete(
                deployment_pb.name, deployment_pb.namespace
            )
        return self.get(namespace=deployment_pb.namespace, name=deployment_pb.name)

    def apply(self, deployment_info, wait):
        from bentoml.yatai.validator import validate_deployment_pb

        if isinstance(deployment_info, dict):
            deployment_pb = deployment_dict_to_pb(deployment_info)
        elif isinstance(deployment_info, str):
            deployment_pb = deployment_yaml_string_to_pb(deployment_info)
        elif isinstance(deployment_info, Deployment):
            deployment_pb = deployment_info
        else:
            raise YataiDeploymentException(
                'Unexpected argument type, expect deployment info to be str in yaml '
                'format or a dict or a deployment protobuf obj, instead got: {}'.format(
                    str(type(deployment_info))
                )
            )

        validation_errors = validate_deployment_pb(deployment_pb)
        if validation_errors:
            raise YataiDeploymentException(
                f'Failed to validate deployment {deployment_pb.name}: '
                f'{validation_errors}'
            )

        apply_result = self.yatai_service.ApplyDeployment(
            ApplyDeploymentRequest(deployment=deployment_pb)
        )
        if apply_result.status.status_code != status_pb2.Status.OK:
            error_code, error_message = status_pb_to_error_code_and_message(
                apply_result.status
            )
            raise YataiDeploymentException(f'{error_code}:{error_message}')
        if wait:
            self._wait_deployment_action_complete(
                deployment_pb.name, deployment_pb.namespace
            )
        return self.get(namespace=deployment_pb.namespace, name=deployment_pb.name)

    def _wait_deployment_action_complete(self, name, namespace):
        start_time = time.time()
        while (time.time() - start_time) < WAIT_TIMEOUT_LIMIT:
            result = self.describe(namespace=namespace, name=name)
            if (
                result.status.status_code == status_pb2.Status.OK
                and result.state.state is DeploymentState.PENDING
            ):
                time.sleep(WAIT_TIME)
                continue
            else:
                break
        return result

    def create_sagemaker_deployment(
        self,
        name,
        bento_name,
        bento_version,
        api_name,
        instance_type,
        instance_count,
        timeout,
        num_of_gunicorn_workers_per_instance=None,
        region=None,
        namespace=None,
        labels=None,
        annotations=None,
        wait=None,
        data_capture_s3_prefix=None,
        data_capture_sample_percent=None,
    ):
        """Create SageMaker deployment

        Args:
            name:
            bento_name:
            bento_version:
            api_name:
            instance_type:
            instance_count:
            timeout:
            num_of_gunicorn_workers_per_instance:
            region:
            namespace:
            labels:
            annotations:
            wait:
            data_capture_s3_prefix:
            data_capture_sample_percent:

        Returns:
            ApplyDeploymentResponse

        Raises:
            BentoMLException
        """
        deployment_pb = Deployment(
            name=name, namespace=namespace, labels=labels, annotations=annotations
        )
        deployment_pb.spec.bento_name = bento_name
        deployment_pb.spec.bento_version = bento_version
        deployment_pb.spec.operator = DeploymentSpec.AWS_SAGEMAKER
        deployment_pb.spec.sagemaker_operator_config.api_name = api_name
        deployment_pb.spec.sagemaker_operator_config.instance_count = instance_count
        deployment_pb.spec.sagemaker_operator_config.instance_type = instance_type
        deployment_pb.spec.sagemaker_operator_config.timeout = timeout

        if data_capture_s3_prefix:
            deployment_pb.spec.sagemaker_operator_config.data_capture_s3_prefix = (
                data_capture_s3_prefix
            )
        if data_capture_sample_percent:
            deployment_pb.spec.sagemaker_operator_config.data_capture_sample_percent = (
                data_capture_sample_percent
            )

        if region:
            deployment_pb.spec.sagemaker_operator_config.region = region
        if num_of_gunicorn_workers_per_instance:
            deployment_pb.spec.sagemaker_operator_config.num_of_gunicorn_workers_per_instance = (  # noqa E501
                num_of_gunicorn_workers_per_instance
            )

        return self.create(deployment_pb, wait)

    def update_sagemaker_deployment(
        self,
        deployment_name,
        namespace=None,
        api_name=None,
        instance_type=None,
        instance_count=None,
        timeout=None,
        num_of_gunicorn_workers_per_instance=None,
        bento_name=None,
        bento_version=None,
        wait=None,
        data_capture_s3_prefix=None,
        data_capture_sample_percent=None,
    ):
        """ Update current sagemaker deployment

        Args:
            namespace:
            deployment_name:
            api_name:
            instance_type:
            instance_count:
            timeout:
            num_of_gunicorn_workers_per_instance:
            bento_name:
            bento_version:
            wait:
            data_capture_s3_prefix:
            data_capture_sample_percent:

        Returns:
            Protobuf message

        Raises:
             BentoMLException
        """

        get_deployment_result = self.get(namespace, deployment_name)
        if get_deployment_result.status.status_code != status_pb2.Status.OK:
            get_deployment_status = get_deployment_result.status
            raise BentoMLException(
                f'Failed to retrieve current deployment {deployment_name} in '
                f'{namespace}. '
                f'{status_pb2.Status.Code.Name(get_deployment_status.status_code)}'
                f':{get_deployment_status.error_message}'
            )

        deployment_pb = get_deployment_result.deployment
        if api_name:
            deployment_pb.spec.sagemaker_operator_config.api_name = api_name
        if instance_type:
            deployment_pb.spec.sagemaker_operator_config.instance_type = instance_type
        if instance_count:
            deployment_pb.spec.sagemaker_operator_config.instance_count = instance_count
        if num_of_gunicorn_workers_per_instance:
            deployment_pb.spec.sagemaker_operator_config.num_of_gunicorn_workers_per_instance = (  # noqa E501
                num_of_gunicorn_workers_per_instance
            )
        if timeout:
            deployment_pb.spec.sagemaker_operator_config.timeout = timeout
        if bento_name:
            deployment_pb.spec.bento_name = bento_name
        if bento_version:
            deployment_pb.spec.bento_version = bento_version
        if data_capture_s3_prefix:
            deployment_pb.spec.sagemaker_operator_config.data_capture_s3_prefix = (
                data_capture_s3_prefix
            )
        if data_capture_sample_percent:
            deployment_pb.spec.sagemaker_operator_config.data_capture_sample_percent = (
                data_capture_sample_percent
            )

        logger.debug(
            'Updated configuration for sagemaker deployment %s', deployment_pb.name
        )

        return self.apply(deployment_pb, wait)

    def list_sagemaker_deployments(
        self,
        limit=None,
        offset=None,
        labels=None,
        namespace=None,
        is_all_namespaces=False,
        order_by=None,
        ascending_order=None,
    ):
        list_result = self.list(
            limit=limit,
            offset=offset,
            labels=labels,
            namespace=namespace,
            is_all_namespaces=is_all_namespaces,
            operator=DeploymentSpec.AWS_SAGEMAKER,
            order_by=order_by,
            ascending_order=ascending_order,
        )
        if list_result.status.status_code != status_pb2.Status.OK:
            return list_result

        sagemaker_deployments = [
            deployment
            for deployment in list_result.deployments
            if deployment.spec.operator == DeploymentSpec.AWS_SAGEMAKER
        ]
        del list_result.deployments[:]
        list_result.deployments.extend(sagemaker_deployments)
        return list_result

    def create_ec2_deployment(
        self,
        name,
        namespace,
        bento_name,
        bento_version,
        region,
        min_size,
        desired_capacity,
        max_size,
        instance_type,
        ami_id,
        wait=None,
    ):

        deployment_pb = Deployment(name=name, namespace=namespace)
        deployment_pb.spec.bento_name = bento_name
        deployment_pb.spec.bento_version = bento_version
        if region:
            deployment_pb.spec.aws_ec2_operator_config.region = region
        deployment_pb.spec.operator = DeploymentSpec.AWS_EC2
        deployment_pb.spec.aws_ec2_operator_config.autoscale_min_size = min_size
        deployment_pb.spec.aws_ec2_operator_config.autoscale_desired_capacity = (
            desired_capacity
        )
        deployment_pb.spec.aws_ec2_operator_config.autoscale_max_size = max_size
        deployment_pb.spec.aws_ec2_operator_config.instance_type = instance_type
        deployment_pb.spec.aws_ec2_operator_config.ami_id = ami_id
        return self.create(deployment_pb, wait)

    def update_ec2_deployment(
        self,
        deployment_name,
        bento_name,
        bento_version,
        namespace,
        min_size,
        desired_capacity,
        max_size,
        instance_type,
        ami_id,
        wait,
    ):
        get_deployment_result = self.get(namespace=namespace, name=deployment_name)
        if get_deployment_result.status.status_code != status_pb2.Status.OK:
            error_code = status_pb2.Status.Code.Name(
                get_deployment_result.status.status_code
            )
            error_message = get_deployment_result.status.error_message
            raise BentoMLException(
                f"Failed to retrieve current deployment {deployment_name} "
                f"in {namespace}.  {error_code}:{error_message}"
            )
        # new deloyment info with updated configs
        deployment_pb = get_deployment_result.deployment

        if bento_name:
            deployment_pb.spec.bento_name = bento_name
        if bento_version:
            deployment_pb.spec.bento_version = bento_version

        deployment_pb.spec.aws_ec2_operator_config.autoscale_min_size = min_size
        deployment_pb.spec.aws_ec2_operator_config.autoscale_desired_capacity = (
            desired_capacity
        )
        deployment_pb.spec.aws_ec2_operator_config.autoscale_max_size = max_size
        deployment_pb.spec.aws_ec2_operator_config.instance_type = instance_type
        deployment_pb.spec.aws_ec2_operator_config.ami_id = ami_id

        logger.debug("Updated configuration for Lambda deployment %s", deployment_name)

        return self.apply(deployment_pb, wait)

    def list_ec2_deployments(
        self,
        limit=None,
        offset=None,
        labels=None,
        namespace=None,
        order_by=None,
        ascending_order=None,
        is_all_namespaces=False,
    ):
        return self.list(
            limit=limit,
            offset=offset,
            labels=labels,
            namespace=namespace,
            is_all_namespaces=is_all_namespaces,
            operator=DeploymentSpec.AWS_EC2,
            order_by=order_by,
            ascending_order=ascending_order,
        )

    def create_lambda_deployment(
        self,
        name,
        bento_name,
        bento_version,
        memory_size,
        timeout,
        api_name=None,
        region=None,
        namespace=None,
        labels=None,
        annotations=None,
        wait=None,
    ):
        """Create Lambda deployment

        Args:
            name:
            bento_name:
            bento_version:
            memory_size:
            timeout:
            api_name:
            region:
            namespace:
            labels:
            annotations:
            wait:

        Returns:
            ApplyDeploymentResponse: status, deployment

        Raises:
            BentoMLException

        """
        deployment_pb = Deployment(
            name=name, namespace=namespace, labels=labels, annotations=annotations
        )
        deployment_pb.spec.bento_name = bento_name
        deployment_pb.spec.bento_version = bento_version
        deployment_pb.spec.operator = DeploymentSpec.AWS_LAMBDA
        deployment_pb.spec.aws_lambda_operator_config.memory_size = memory_size
        deployment_pb.spec.aws_lambda_operator_config.timeout = timeout
        if api_name:
            deployment_pb.spec.aws_lambda_operator_config.api_name = api_name
        if region:
            deployment_pb.spec.aws_lambda_operator_config.region = region
        return self.create(deployment_pb, wait)

    def update_lambda_deployment(
        self,
        deployment_name,
        namespace=None,
        bento_name=None,
        bento_version=None,
        memory_size=None,
        timeout=None,
        wait=None,
    ):
        get_deployment_result = self.get(namespace=namespace, name=deployment_name)
        if get_deployment_result.status.status_code != status_pb2.Status.OK:
            error_code = status_pb2.Status.Code.Name(
                get_deployment_result.status.status_code
            )
            error_message = status_pb2.status.error_message
            raise BentoMLException(
                f'Failed to retrieve current deployment {deployment_name} '
                f'in {namespace}.  {error_code}:{error_message}'
            )
        deployment_pb = get_deployment_result.deployment
        if bento_name:
            deployment_pb.spec.bento_name = bento_name
        if bento_version:
            deployment_pb.spec.bento_version = bento_version
        if memory_size:
            deployment_pb.spec.aws_lambda_operator_config.memory_size = memory_size
        if timeout:
            deployment_pb.spec.aws_lambda_operator_config.timeout = timeout
        logger.debug('Updated configuration for Lambda deployment %s', deployment_name)

        return self.apply(deployment_pb, wait)

    def list_lambda_deployments(
        self,
        limit=None,
        offset=None,
        labels=None,
        namespace=None,
        is_all_namespaces=False,
        order_by=None,
        ascending_order=None,
    ):
        return self.list(
            limit=limit,
            offset=offset,
            labels=labels,
            namespace=namespace,
            is_all_namespaces=is_all_namespaces,
            operator=DeploymentSpec.AWS_LAMBDA,
            order_by=order_by,
            ascending_order=ascending_order,
        )

    def create_azure_functions_deployment(
        self,
        name,
        bento_name,
        bento_version,
        location,
        premium_plan_sku=None,
        min_instances=None,
        max_burst=None,
        function_auth_level=None,
        namespace=None,
        labels=None,
        annotations=None,
        wait=None,
    ):
        deployment_pb = Deployment(
            name=name, namespace=namespace, labels=labels, annotations=annotations
        )

        deployment_pb.spec.bento_name = bento_name
        deployment_pb.spec.bento_version = bento_version
        deployment_pb.spec.operator = DeploymentSpec.AZURE_FUNCTIONS
        deployment_pb.spec.azure_functions_operator_config.location = location
        deployment_pb.spec.azure_functions_operator_config.premium_plan_sku = (
            premium_plan_sku
        )
        deployment_pb.spec.azure_functions_operator_config.min_instances = min_instances
        deployment_pb.spec.azure_functions_operator_config.function_auth_level = (
            function_auth_level
        )
        deployment_pb.spec.azure_functions_operator_config.max_burst = max_burst

        return self.create(deployment_pb, wait)

    def update_azure_functions_deployment(
        self,
        deployment_name,
        bento_name=None,
        bento_version=None,
        max_burst=None,
        min_instances=None,
        premium_plan_sku=None,
        namespace=None,
        wait=None,
    ):
        get_deployment_result = self.get(namespace=namespace, name=deployment_name)
        if get_deployment_result.status.status_code != status_pb2.Status.OK:
            error_code = status_pb2.Status.Code.Name(
                get_deployment_result.status.status_code
            )
            error_message = status_pb2.status.error_message
            raise BentoMLException(
                f'Failed to retrieve current deployment {deployment_name} in '
                f'{namespace}. {error_code}:{error_message}'
            )
        deployment_pb = get_deployment_result.deployment
        if bento_name:
            deployment_pb.spec.bento_name = bento_name
        if bento_version:
            deployment_pb.spec.bento_version = bento_version
        if max_burst:
            deployment_pb.spec.azure_functions_operator_config.max_burst = max_burst
        if min_instances:
            deployment_pb.spec.azure_functions_operator_config.min_instances = (
                min_instances
            )
        if premium_plan_sku:
            deployment_pb.spec.azure_functions_operator_config.premium_plan_sku = (
                premium_plan_sku
            )
        return self.apply(deployment_pb, wait)

    def list_azure_functions_deployments(
        self,
        limit=None,
        offset=None,
        labels=None,
        namespace=None,
        is_all_namespaces=False,
        order_by=None,
        ascending_order=None,
    ):
        return self.list(
            limit=limit,
            offset=offset,
            labels=labels,
            namespace=namespace,
            is_all_namespaces=is_all_namespaces,
            operator=DeploymentSpec.AZURE_FUNCTIONS,
            order_by=order_by,
            ascending_order=ascending_order,
        )



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ._version import get_versions

__version__ = get_versions()['version']
del get_versions

from bentoml.configuration import inject_dependencies
from bentoml.utils.log import configure_logging

# Inject dependencies and configurations
inject_dependencies()

# Configuring logging properly before loading other modules
configure_logging()

from bentoml.saved_bundle import load_from_dir, save_to_dir  # noqa: E402
from bentoml.service import (  # noqa: E402
    BentoService,
    api_decorator as api,
    env_decorator as env,
    web_static_content_decorator as web_static_content,
    artifacts_decorator as artifacts,
    ver_decorator as ver,
    save,
)

from bentoml.cli import create_bentoml_cli

commandline_interface = create_bentoml_cli()

load = load_from_dir

__all__ = [
    "__version__",
    "api",
    "artifacts",
    "env",
    "web_static_content",
    "ver",
    "BentoService",
    # backward compatible
    "load",
    "load_from_dir",
    "save",
    "save_to_dir",
    "handlers",
    "adapters",
    "artifact",
]



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from bentoml.cli.aws_lambda import get_aws_lambda_sub_command
from bentoml.cli.aws_sagemaker import get_aws_sagemaker_sub_command
from bentoml.cli.azure_functions import get_azure_functions_sub_command
from bentoml.cli.aws_ec2 import get_aws_ec2_sub_command
from bentoml.cli.bento_management import add_bento_sub_command
from bentoml.cli.bento_service import create_bento_service_cli
from bentoml.cli.yatai_service import add_yatai_service_sub_command
from bentoml.cli.deployment import get_deployment_sub_command


def create_bentoml_cli():
    # pylint: disable=unused-variable

    _cli = create_bento_service_cli()

    # Commands created here aren't mean to be used from generated BentoService CLI when
    # installed as PyPI package. The are only used as part of BentoML cli command.

    aws_sagemaker_sub_command = get_aws_sagemaker_sub_command()
    aws_lambda_sub_command = get_aws_lambda_sub_command()
    deployment_sub_command = get_deployment_sub_command()
    azure_function_sub_command = get_azure_functions_sub_command()
    aws_ec2_sub_command = get_aws_ec2_sub_command()
    add_bento_sub_command(_cli)
    add_yatai_service_sub_command(_cli)
    _cli.add_command(aws_sagemaker_sub_command)
    _cli.add_command(aws_lambda_sub_command)
    _cli.add_command(aws_ec2_sub_command)
    _cli.add_command(azure_function_sub_command)
    _cli.add_command(deployment_sub_command)

    return _cli


if __name__ == "__main__":
    from bentoml import commandline_interface

    commandline_interface()  # pylint: disable=no-value-for-parameter



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import shutil
import uuid
import logging
from pathlib import Path

import boto3

from bentoml.utils.ruamel_yaml import YAML
from bentoml.exceptions import (
    BentoMLException,
    InvalidArgument,
    YataiDeploymentException,
)
from bentoml.saved_bundle import loader
from bentoml.utils import status_pb_to_error_code_and_message
from bentoml.utils.s3 import create_s3_bucket_if_not_exists
from bentoml.utils.tempdir import TempDirectory
from bentoml.yatai.deployment.aws_utils import (
    validate_sam_template,
    FAILED_CLOUDFORMATION_STACK_STATUS,
    cleanup_s3_bucket_if_exist,
)
from bentoml.yatai.deployment.aws_lambda.utils import (
    init_sam_project,
    lambda_deploy,
    lambda_package,
    reduce_bundle_size_and_upload_extra_resources_to_s3,
    total_file_or_directory_size,
    LAMBDA_FUNCTION_LIMIT,
    LAMBDA_FUNCTION_MAX_LIMIT,
)

from bentoml.yatai.deployment.operator import DeploymentOperatorBase
from bentoml.yatai.deployment.utils import (
    raise_if_api_names_not_found_in_bento_service_metadata,
)
from bentoml.yatai.deployment.aws_utils import (
    generate_aws_compatible_string,
    get_default_aws_region,
    ensure_sam_available_or_raise,
)
from bentoml.yatai.deployment.docker_utils import ensure_docker_available_or_raise
from bentoml.yatai.proto import status_pb2
from bentoml.yatai.proto.deployment_pb2 import (
    DeploymentState,
    ApplyDeploymentResponse,
    DeleteDeploymentResponse,
    DescribeDeploymentResponse,
)
from bentoml.yatai.proto.repository_pb2 import GetBentoRequest, BentoUri
from bentoml.yatai.status import Status


logger = logging.getLogger(__name__)


def _create_aws_lambda_cloudformation_template_file(
    project_dir,
    namespace,
    deployment_name,
    deployment_path_prefix,
    api_names,
    bento_service_name,
    s3_bucket_name,
    py_runtime,
    memory_size,
    timeout,
):
    template_file_path = os.path.join(project_dir, 'template.yaml')
    yaml = YAML()
    sam_config = {
        'AWSTemplateFormatVersion': '2010-09-09',
        'Transform': 'AWS::Serverless-2016-10-31',
        'Globals': {
            'Function': {'Timeout': timeout, 'Runtime': py_runtime},
            'Api': {
                'BinaryMediaTypes': ['image~1*'],
                'Cors': "'*'",
                'Auth': {
                    'ApiKeyRequired': False,
                    'DefaultAuthorizer': 'NONE',
                    'AddDefaultAuthorizerToCorsPreflight': False,
                },
            },
        },
        'Resources': {},
        'Outputs': {
            'S3Bucket': {
                'Value': s3_bucket_name,
                'Description': 'S3 Bucket for saving artifacts and lambda bundle',
            }
        },
    }
    for api_name in api_names:
        sam_config['Resources'][api_name] = {
            'Type': 'AWS::Serverless::Function',
            'Properties': {
                'Runtime': py_runtime,
                'CodeUri': deployment_name + '/',
                'Handler': 'app.{}'.format(api_name),
                'FunctionName': f'{namespace}-{deployment_name}-{api_name}',
                'Timeout': timeout,
                'MemorySize': memory_size,
                'Events': {
                    'Api': {
                        'Type': 'Api',
                        'Properties': {
                            'Path': '/{}'.format(api_name),
                            'Method': 'post',
                        },
                    }
                },
                'Policies': [{'S3ReadPolicy': {'BucketName': s3_bucket_name}}],
                'Environment': {
                    'Variables': {
                        'BENTOML_BENTO_SERVICE_NAME': bento_service_name,
                        'BENTOML_API_NAME': api_name,
                        'BENTOML_S3_BUCKET': s3_bucket_name,
                        'BENTOML_DEPLOYMENT_PATH_PREFIX': deployment_path_prefix,
                    }
                },
            },
        }

    yaml.dump(sam_config, Path(template_file_path))

    # We add Outputs section separately, because the value should not
    # have "'" around !Sub
    with open(template_file_path, 'a') as f:
        f.write(
            """\
  EndpointUrl:
    Value: !Sub "https://${ServerlessRestApi}.execute-api.${AWS::Region}.\
amazonaws.com/Prod"
    Description: URL for endpoint
"""
        )
    return template_file_path


def _deploy_lambda_function(
    deployment_pb,
    bento_service_metadata,
    deployment_spec,
    lambda_s3_bucket,
    lambda_deployment_config,
    bento_path,
):
    deployment_path_prefix = os.path.join(deployment_pb.namespace, deployment_pb.name)

    py_major, py_minor, _ = bento_service_metadata.env.python_version.split('.')
    if py_major != '3':
        raise BentoMLException('Python 2 is not supported for Lambda Deployment')
    python_runtime = 'python{}.{}'.format(py_major, py_minor)

    artifact_types = [item.artifact_type for item in bento_service_metadata.artifacts]
    if any(
        i in ['TensorflowSavedModelArtifact', 'KerasModelArtifact']
        for i in artifact_types
    ) and (py_major, py_minor) != ('3', '6'):
        raise BentoMLException(
            'AWS Lambda Deployment only supports BentoML services'
            'built with Python 3.6.x. To fix this, repack your'
            'service with the right Python version'
            '(hint: pyenv/anaconda) and try again'
        )

    api_names = (
        [lambda_deployment_config.api_name]
        if lambda_deployment_config.api_name
        else [api.name for api in bento_service_metadata.apis]
    )

    raise_if_api_names_not_found_in_bento_service_metadata(
        bento_service_metadata, api_names
    )

    with TempDirectory() as lambda_project_dir:
        logger.debug(
            'Generating cloudformation template.yaml for lambda project at %s',
            lambda_project_dir,
        )
        template_file_path = _create_aws_lambda_cloudformation_template_file(
            project_dir=lambda_project_dir,
            namespace=deployment_pb.namespace,
            deployment_name=deployment_pb.name,
            deployment_path_prefix=deployment_path_prefix,
            api_names=api_names,
            bento_service_name=deployment_spec.bento_name,
            s3_bucket_name=lambda_s3_bucket,
            py_runtime=python_runtime,
            memory_size=lambda_deployment_config.memory_size,
            timeout=lambda_deployment_config.timeout,
        )
        logger.debug('Validating generated template.yaml')
        validate_sam_template(
            template_file_path, lambda_deployment_config.region, lambda_project_dir,
        )
        logger.debug(
            'Initializing lambda project in directory: %s ...', lambda_project_dir,
        )
        init_sam_project(
            lambda_project_dir,
            bento_path,
            deployment_pb.name,
            deployment_spec.bento_name,
            api_names,
            aws_region=lambda_deployment_config.region,
        )
        for api_name in api_names:
            build_directory = os.path.join(
                lambda_project_dir, '.aws-sam', 'build', api_name
            )
            logger.debug(
                'Checking is function "%s" bundle under lambda size ' 'limit', api_name,
            )
            # Since we only use s3 get object in lambda function, and
            # lambda function pack their own boto3/botocore modules,
            # we will just delete those modules from function bundle
            # directory
            delete_list = ['boto3', 'botocore']
            for name in delete_list:
                logger.debug('Remove module "%s" from build directory', name)
                shutil.rmtree(os.path.join(build_directory, name))
            total_build_dir_size = total_file_or_directory_size(build_directory)
            if total_build_dir_size > LAMBDA_FUNCTION_MAX_LIMIT:
                raise BentoMLException(
                    'Build function size is over 700MB, max size '
                    'capable for AWS Lambda function'
                )
            if total_build_dir_size >= LAMBDA_FUNCTION_LIMIT:
                logger.debug(
                    'Function %s is over lambda size limit, attempting ' 'reduce it',
                    api_name,
                )
                reduce_bundle_size_and_upload_extra_resources_to_s3(
                    build_directory=build_directory,
                    region=lambda_deployment_config.region,
                    s3_bucket=lambda_s3_bucket,
                    deployment_prefix=deployment_path_prefix,
                    function_name=api_name,
                    lambda_project_dir=lambda_project_dir,
                )
            else:
                logger.debug(
                    'Function bundle is within Lambda limit, removing '
                    'download_extra_resources.py file from function bundle'
                )
                os.remove(os.path.join(build_directory, 'download_extra_resources.py'))
        logger.info('Packaging AWS Lambda project at %s ...', lambda_project_dir)
        lambda_package(
            lambda_project_dir,
            lambda_deployment_config.region,
            lambda_s3_bucket,
            deployment_path_prefix,
        )
        logger.info('Deploying lambda project')
        stack_name = generate_aws_compatible_string(
            deployment_pb.namespace + '-' + deployment_pb.name
        )
        lambda_deploy(
            lambda_project_dir, lambda_deployment_config.region, stack_name=stack_name,
        )


class AwsLambdaDeploymentOperator(DeploymentOperatorBase):
    def __init__(self, yatai_service):
        super(AwsLambdaDeploymentOperator, self).__init__(yatai_service)
        ensure_docker_available_or_raise()
        ensure_sam_available_or_raise()

    def add(self, deployment_pb):
        try:
            deployment_spec = deployment_pb.spec
            deployment_spec.aws_lambda_operator_config.region = (
                deployment_spec.aws_lambda_operator_config.region
                or get_default_aws_region()
            )
            if not deployment_spec.aws_lambda_operator_config.region:
                raise InvalidArgument('AWS region is missing')

            bento_pb = self.yatai_service.GetBento(
                GetBentoRequest(
                    bento_name=deployment_spec.bento_name,
                    bento_version=deployment_spec.bento_version,
                )
            )
            if bento_pb.bento.uri.type not in (BentoUri.LOCAL, BentoUri.S3):
                raise BentoMLException(
                    'BentoML currently not support {} repository'.format(
                        BentoUri.StorageType.Name(bento_pb.bento.uri.type)
                    )
                )

            return self._add(deployment_pb, bento_pb, bento_pb.bento.uri.uri)
        except BentoMLException as error:
            deployment_pb.state.state = DeploymentState.ERROR
            deployment_pb.state.error_message = f'Error: {str(error)}'
            return ApplyDeploymentResponse(
                status=error.status_proto, deployment=deployment_pb
            )

    def _add(self, deployment_pb, bento_pb, bento_path):
        if loader._is_remote_path(bento_path):
            with loader._resolve_remote_bundle_path(bento_path) as local_path:
                return self._add(deployment_pb, bento_pb, local_path)

        deployment_spec = deployment_pb.spec
        lambda_deployment_config = deployment_spec.aws_lambda_operator_config
        bento_service_metadata = bento_pb.bento.bento_service_metadata
        lambda_s3_bucket = generate_aws_compatible_string(
            'btml-{namespace}-{name}-{random_string}'.format(
                namespace=deployment_pb.namespace,
                name=deployment_pb.name,
                random_string=uuid.uuid4().hex[:6].lower(),
            )
        )
        try:
            create_s3_bucket_if_not_exists(
                lambda_s3_bucket, lambda_deployment_config.region
            )
            _deploy_lambda_function(
                deployment_pb=deployment_pb,
                bento_service_metadata=bento_service_metadata,
                deployment_spec=deployment_spec,
                lambda_s3_bucket=lambda_s3_bucket,
                lambda_deployment_config=lambda_deployment_config,
                bento_path=bento_path,
            )
            return ApplyDeploymentResponse(status=Status.OK(), deployment=deployment_pb)
        except BentoMLException as error:
            if lambda_s3_bucket and lambda_deployment_config:
                cleanup_s3_bucket_if_exist(
                    lambda_s3_bucket, lambda_deployment_config.region
                )
            raise error

    def update(self, deployment_pb, previous_deployment):
        try:
            deployment_spec = deployment_pb.spec
            bento_pb = self.yatai_service.GetBento(
                GetBentoRequest(
                    bento_name=deployment_spec.bento_name,
                    bento_version=deployment_spec.bento_version,
                )
            )
            if bento_pb.bento.uri.type not in (BentoUri.LOCAL, BentoUri.S3):
                raise BentoMLException(
                    'BentoML currently not support {} repository'.format(
                        BentoUri.StorageType.Name(bento_pb.bento.uri.type)
                    )
                )

            return self._update(
                deployment_pb, previous_deployment, bento_pb, bento_pb.bento.uri.uri
            )
        except BentoMLException as error:
            deployment_pb.state.state = DeploymentState.ERROR
            deployment_pb.state.error_message = f'Error: {str(error)}'
            return ApplyDeploymentResponse(
                status=error.status_code, deployment_pb=deployment_pb
            )

    def _update(self, deployment_pb, current_deployment, bento_pb, bento_path):
        if loader._is_remote_path(bento_path):
            with loader._resolve_remote_bundle_path(bento_path) as local_path:
                return self._update(
                    deployment_pb, current_deployment, bento_pb, local_path
                )
        updated_deployment_spec = deployment_pb.spec
        updated_lambda_deployment_config = (
            updated_deployment_spec.aws_lambda_operator_config
        )
        updated_bento_service_metadata = bento_pb.bento.bento_service_metadata
        describe_result = self.describe(deployment_pb)
        if describe_result.status.status_code != status_pb2.Status.OK:
            error_code, error_message = status_pb_to_error_code_and_message(
                describe_result.status
            )
            raise YataiDeploymentException(
                f'Failed fetching Lambda deployment current status - '
                f'{error_code}:{error_message}'
            )
        latest_deployment_state = json.loads(describe_result.state.info_json)
        if 's3_bucket' in latest_deployment_state:
            lambda_s3_bucket = latest_deployment_state['s3_bucket']
        else:
            raise BentoMLException(
                'S3 Bucket is missing in the AWS Lambda deployment, please make sure '
                'it exists and try again'
            )

        _deploy_lambda_function(
            deployment_pb=deployment_pb,
            bento_service_metadata=updated_bento_service_metadata,
            deployment_spec=updated_deployment_spec,
            lambda_s3_bucket=lambda_s3_bucket,
            lambda_deployment_config=updated_lambda_deployment_config,
            bento_path=bento_path,
        )

        return ApplyDeploymentResponse(deployment=deployment_pb, status=Status.OK())

    def delete(self, deployment_pb):
        try:
            logger.debug('Deleting AWS Lambda deployment')

            deployment_spec = deployment_pb.spec
            lambda_deployment_config = deployment_spec.aws_lambda_operator_config
            lambda_deployment_config.region = (
                lambda_deployment_config.region or get_default_aws_region()
            )
            if not lambda_deployment_config.region:
                raise InvalidArgument('AWS region is missing')

            cf_client = boto3.client('cloudformation', lambda_deployment_config.region)
            stack_name = generate_aws_compatible_string(
                deployment_pb.namespace, deployment_pb.name
            )
            if deployment_pb.state.info_json:
                deployment_info_json = json.loads(deployment_pb.state.info_json)
                bucket_name = deployment_info_json.get('s3_bucket')
                if bucket_name:
                    cleanup_s3_bucket_if_exist(
                        bucket_name, lambda_deployment_config.region
                    )

            logger.debug(
                'Deleting AWS CloudFormation: %s that includes Lambda function '
                'and related resources',
                stack_name,
            )
            cf_client.delete_stack(StackName=stack_name)
            return DeleteDeploymentResponse(status=Status.OK())

        except BentoMLException as error:
            return DeleteDeploymentResponse(status=error.status_proto)

    def describe(self, deployment_pb):
        try:
            deployment_spec = deployment_pb.spec
            lambda_deployment_config = deployment_spec.aws_lambda_operator_config
            lambda_deployment_config.region = (
                lambda_deployment_config.region or get_default_aws_region()
            )
            if not lambda_deployment_config.region:
                raise InvalidArgument('AWS region is missing')

            bento_pb = self.yatai_service.GetBento(
                GetBentoRequest(
                    bento_name=deployment_spec.bento_name,
                    bento_version=deployment_spec.bento_version,
                )
            )
            bento_service_metadata = bento_pb.bento.bento_service_metadata
            api_names = (
                [lambda_deployment_config.api_name]
                if lambda_deployment_config.api_name
                else [api.name for api in bento_service_metadata.apis]
            )

            try:
                cf_client = boto3.client(
                    'cloudformation', lambda_deployment_config.region
                )
                stack_name = generate_aws_compatible_string(
                    '{ns}-{name}'.format(
                        ns=deployment_pb.namespace, name=deployment_pb.name
                    )
                )
                cloud_formation_stack_result = cf_client.describe_stacks(
                    StackName=stack_name
                )
                stack_result = cloud_formation_stack_result.get('Stacks')[0]
                # https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/\
                # using-cfn-describing-stacks.html
                success_status = ['CREATE_COMPLETE', 'UPDATE_COMPLETE']
                if stack_result['StackStatus'] in success_status:
                    if stack_result.get('Outputs'):
                        outputs = stack_result['Outputs']
                    else:
                        return DescribeDeploymentResponse(
                            status=Status.ABORTED('"Outputs" field is not present'),
                            state=DeploymentState(
                                state=DeploymentState.ERROR,
                                error_message='"Outputs" field is not present',
                            ),
                        )
                elif stack_result['StackStatus'] in FAILED_CLOUDFORMATION_STACK_STATUS:
                    state = DeploymentState(state=DeploymentState.FAILED)
                    state.timestamp.GetCurrentTime()
                    return DescribeDeploymentResponse(status=Status.OK(), state=state)
                else:
                    state = DeploymentState(state=DeploymentState.PENDING)
                    state.timestamp.GetCurrentTime()
                    return DescribeDeploymentResponse(status=Status.OK(), state=state)
            except Exception as error:  # pylint: disable=broad-except
                state = DeploymentState(
                    state=DeploymentState.ERROR, error_message=str(error)
                )
                state.timestamp.GetCurrentTime()
                return DescribeDeploymentResponse(
                    status=Status.INTERNAL(str(error)), state=state
                )
            outputs = {o['OutputKey']: o['OutputValue'] for o in outputs}
            info_json = {}

            if 'EndpointUrl' in outputs:
                info_json['endpoints'] = [
                    outputs['EndpointUrl'] + '/' + api_name for api_name in api_names
                ]
            if 'S3Bucket' in outputs:
                info_json['s3_bucket'] = outputs['S3Bucket']

            state = DeploymentState(
                state=DeploymentState.RUNNING, info_json=json.dumps(info_json)
            )
            state.timestamp.GetCurrentTime()
            return DescribeDeploymentResponse(status=Status.OK(), state=state)
        except BentoMLException as error:
            return DescribeDeploymentResponse(status=error.status_proto)



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import traceback
from typing import Iterable, Sequence, Tuple

from bentoml.adapters.string_input import StringInput
from bentoml.types import InferenceTask, JsonSerializable

ApiFuncArgs = Tuple[
    Sequence[JsonSerializable],
]


class JsonInput(StringInput):
    """JsonInput parses REST API request or CLI command into parsed_jsons(a list of
    json serializable object in python) and pass down to user defined API function

    Examples
    ----------
    Example services:

    Use JsonInput with ``batch=True``

    .. code-block:: python

        from typing import List
        from bentoml.types import JsonSerializable

        @bentoml.api(input=JsonInput(), batch=True)
        def predict(self, parsed_json_list: List[JsonSerializable]):
            results = self.artifacts.classifier([j['text'] for j in parsed_json_list])
            return results

    OR use JsonInput with ``batch=False`` (the default)

    .. code-block:: python

        @bentoml.api(input=JsonInput())
        def predict(self, parsed_json):
            results = self.artifacts.classifier([parsed_json['text']])
            return results[0]

    For client prediction request, it is the same for both batch and non-batch API,
    the request should contain only one single input item::

        curl -i \\
          --header "Content-Type: application/json" \\
          --request POST \\
          --data '{"text": "best movie ever"}' \\
          localhost:5000/predict
    """

    def extract_user_func_args(
        self, tasks: Iterable[InferenceTask[str]]
    ) -> ApiFuncArgs:
        json_inputs = []
        for task in tasks:
            try:
                parsed_json = json.loads(task.data)
                json_inputs.append(parsed_json)
            except json.JSONDecodeError:
                task.discard(http_status=400, err_msg="Not a valid JSON format")
            except Exception:  # pylint: disable=broad-except
                err = traceback.format_exc()
                task.discard(http_status=500, err_msg=f"Internal Server Error: {err}")
        return (json_inputs,)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

import bentoml.yatai.proto.deployment_pb2 as deployment__pb2
from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
import bentoml.yatai.proto.repository_pb2 as repository__pb2
import bentoml.yatai.proto.yatai_service_pb2 as yatai__service__pb2


class YataiStub(object):
    """Yatai RPC Server

    A stateful service that provides a complete BentoML model management
    and model serving/deployment workflow

    It provides two sets of APIs:
    Bento Repository: Manages saved Bento files, and making them available
    for serving in production environments
    Serving Deployment: Deploys saved Bento to a varity of different cloud
    platforms, track deploym//ent status, set up logging
    monitoring for your model serving workloads, and
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.HealthCheck = channel.unary_unary(
                '/bentoml.Yatai/HealthCheck',
                request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
                response_deserializer=yatai__service__pb2.HealthCheckResponse.FromString,
                )
        self.GetYataiServiceVersion = channel.unary_unary(
                '/bentoml.Yatai/GetYataiServiceVersion',
                request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
                response_deserializer=yatai__service__pb2.GetYataiServiceVersionResponse.FromString,
                )
        self.ApplyDeployment = channel.unary_unary(
                '/bentoml.Yatai/ApplyDeployment',
                request_serializer=deployment__pb2.ApplyDeploymentRequest.SerializeToString,
                response_deserializer=deployment__pb2.ApplyDeploymentResponse.FromString,
                )
        self.DeleteDeployment = channel.unary_unary(
                '/bentoml.Yatai/DeleteDeployment',
                request_serializer=deployment__pb2.DeleteDeploymentRequest.SerializeToString,
                response_deserializer=deployment__pb2.DeleteDeploymentResponse.FromString,
                )
        self.GetDeployment = channel.unary_unary(
                '/bentoml.Yatai/GetDeployment',
                request_serializer=deployment__pb2.GetDeploymentRequest.SerializeToString,
                response_deserializer=deployment__pb2.GetDeploymentResponse.FromString,
                )
        self.DescribeDeployment = channel.unary_unary(
                '/bentoml.Yatai/DescribeDeployment',
                request_serializer=deployment__pb2.DescribeDeploymentRequest.SerializeToString,
                response_deserializer=deployment__pb2.DescribeDeploymentResponse.FromString,
                )
        self.ListDeployments = channel.unary_unary(
                '/bentoml.Yatai/ListDeployments',
                request_serializer=deployment__pb2.ListDeploymentsRequest.SerializeToString,
                response_deserializer=deployment__pb2.ListDeploymentsResponse.FromString,
                )
        self.AddBento = channel.unary_unary(
                '/bentoml.Yatai/AddBento',
                request_serializer=repository__pb2.AddBentoRequest.SerializeToString,
                response_deserializer=repository__pb2.AddBentoResponse.FromString,
                )
        self.UpdateBento = channel.unary_unary(
                '/bentoml.Yatai/UpdateBento',
                request_serializer=repository__pb2.UpdateBentoRequest.SerializeToString,
                response_deserializer=repository__pb2.UpdateBentoResponse.FromString,
                )
        self.GetBento = channel.unary_unary(
                '/bentoml.Yatai/GetBento',
                request_serializer=repository__pb2.GetBentoRequest.SerializeToString,
                response_deserializer=repository__pb2.GetBentoResponse.FromString,
                )
        self.DangerouslyDeleteBento = channel.unary_unary(
                '/bentoml.Yatai/DangerouslyDeleteBento',
                request_serializer=repository__pb2.DangerouslyDeleteBentoRequest.SerializeToString,
                response_deserializer=repository__pb2.DangerouslyDeleteBentoResponse.FromString,
                )
        self.ListBento = channel.unary_unary(
                '/bentoml.Yatai/ListBento',
                request_serializer=repository__pb2.ListBentoRequest.SerializeToString,
                response_deserializer=repository__pb2.ListBentoResponse.FromString,
                )
        self.ContainerizeBento = channel.unary_unary(
                '/bentoml.Yatai/ContainerizeBento',
                request_serializer=repository__pb2.ContainerizeBentoRequest.SerializeToString,
                response_deserializer=repository__pb2.ContainerizeBentoResponse.FromString,
                )
        self.UploadBento = channel.stream_unary(
                '/bentoml.Yatai/UploadBento',
                request_serializer=repository__pb2.UploadBentoRequest.SerializeToString,
                response_deserializer=repository__pb2.UploadBentoResponse.FromString,
                )
        self.DownloadBento = channel.unary_stream(
                '/bentoml.Yatai/DownloadBento',
                request_serializer=repository__pb2.DownloadBentoRequest.SerializeToString,
                response_deserializer=repository__pb2.DownloadBentoResponse.FromString,
                )


class YataiServicer(object):
    """Yatai RPC Server

    A stateful service that provides a complete BentoML model management
    and model serving/deployment workflow

    It provides two sets of APIs:
    Bento Repository: Manages saved Bento files, and making them available
    for serving in production environments
    Serving Deployment: Deploys saved Bento to a varity of different cloud
    platforms, track deploym//ent status, set up logging
    monitoring for your model serving workloads, and
    """

    def HealthCheck(self, request, context):
        """Health check ping endpoint
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetYataiServiceVersion(self, request, context):
        """Return current service version information
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ApplyDeployment(self, request, context):
        """Create new or update existing deployment
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DeleteDeployment(self, request, context):
        """Delete existing deployment
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetDeployment(self, request, context):
        """Get deployment specification (desired state)
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DescribeDeployment(self, request, context):
        """Get deployment status (current state)
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListDeployments(self, request, context):
        """List active deployments, by default this will return all active deployments
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def AddBento(self, request, context):
        """Add new saved Bento to repository by providing the Bento name and version
        this will return an upload address that allows client to upload the bento files
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def UpdateBento(self, request, context):
        """RPC for updating a previously added Bento's information, including
        the BentoService's Metadata(apis, env, artifacts etc) and the upload status.
        Yatai server expects the client to use this RPC for notifying that, for a
        previously requested AddBento call, what's the uploading progress and when the
        upload is completed
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetBento(self, request, context):
        """Get a file path to the saved Bento files, path must be accessible form client
        machine either through HTTP, FTP, etc
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DangerouslyDeleteBento(self, request, context):
        """Deleting the Bento files that was added to this Yatai server earlier, this may
        break existing deployments or create issues when doing deployment rollback
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListBento(self, request, context):
        """Get a list of Bento that are stored in current repository
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ContainerizeBento(self, request, context):
        """Create a container image from a Bento
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def UploadBento(self, request_iterator, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DownloadBento(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_YataiServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'HealthCheck': grpc.unary_unary_rpc_method_handler(
                    servicer.HealthCheck,
                    request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                    response_serializer=yatai__service__pb2.HealthCheckResponse.SerializeToString,
            ),
            'GetYataiServiceVersion': grpc.unary_unary_rpc_method_handler(
                    servicer.GetYataiServiceVersion,
                    request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                    response_serializer=yatai__service__pb2.GetYataiServiceVersionResponse.SerializeToString,
            ),
            'ApplyDeployment': grpc.unary_unary_rpc_method_handler(
                    servicer.ApplyDeployment,
                    request_deserializer=deployment__pb2.ApplyDeploymentRequest.FromString,
                    response_serializer=deployment__pb2.ApplyDeploymentResponse.SerializeToString,
            ),
            'DeleteDeployment': grpc.unary_unary_rpc_method_handler(
                    servicer.DeleteDeployment,
                    request_deserializer=deployment__pb2.DeleteDeploymentRequest.FromString,
                    response_serializer=deployment__pb2.DeleteDeploymentResponse.SerializeToString,
            ),
            'GetDeployment': grpc.unary_unary_rpc_method_handler(
                    servicer.GetDeployment,
                    request_deserializer=deployment__pb2.GetDeploymentRequest.FromString,
                    response_serializer=deployment__pb2.GetDeploymentResponse.SerializeToString,
            ),
            'DescribeDeployment': grpc.unary_unary_rpc_method_handler(
                    servicer.DescribeDeployment,
                    request_deserializer=deployment__pb2.DescribeDeploymentRequest.FromString,
                    response_serializer=deployment__pb2.DescribeDeploymentResponse.SerializeToString,
            ),
            'ListDeployments': grpc.unary_unary_rpc_method_handler(
                    servicer.ListDeployments,
                    request_deserializer=deployment__pb2.ListDeploymentsRequest.FromString,
                    response_serializer=deployment__pb2.ListDeploymentsResponse.SerializeToString,
            ),
            'AddBento': grpc.unary_unary_rpc_method_handler(
                    servicer.AddBento,
                    request_deserializer=repository__pb2.AddBentoRequest.FromString,
                    response_serializer=repository__pb2.AddBentoResponse.SerializeToString,
            ),
            'UpdateBento': grpc.unary_unary_rpc_method_handler(
                    servicer.UpdateBento,
                    request_deserializer=repository__pb2.UpdateBentoRequest.FromString,
                    response_serializer=repository__pb2.UpdateBentoResponse.SerializeToString,
            ),
            'GetBento': grpc.unary_unary_rpc_method_handler(
                    servicer.GetBento,
                    request_deserializer=repository__pb2.GetBentoRequest.FromString,
                    response_serializer=repository__pb2.GetBentoResponse.SerializeToString,
            ),
            'DangerouslyDeleteBento': grpc.unary_unary_rpc_method_handler(
                    servicer.DangerouslyDeleteBento,
                    request_deserializer=repository__pb2.DangerouslyDeleteBentoRequest.FromString,
                    response_serializer=repository__pb2.DangerouslyDeleteBentoResponse.SerializeToString,
            ),
            'ListBento': grpc.unary_unary_rpc_method_handler(
                    servicer.ListBento,
                    request_deserializer=repository__pb2.ListBentoRequest.FromString,
                    response_serializer=repository__pb2.ListBentoResponse.SerializeToString,
            ),
            'ContainerizeBento': grpc.unary_unary_rpc_method_handler(
                    servicer.ContainerizeBento,
                    request_deserializer=repository__pb2.ContainerizeBentoRequest.FromString,
                    response_serializer=repository__pb2.ContainerizeBentoResponse.SerializeToString,
            ),
            'UploadBento': grpc.stream_unary_rpc_method_handler(
                    servicer.UploadBento,
                    request_deserializer=repository__pb2.UploadBentoRequest.FromString,
                    response_serializer=repository__pb2.UploadBentoResponse.SerializeToString,
            ),
            'DownloadBento': grpc.unary_stream_rpc_method_handler(
                    servicer.DownloadBento,
                    request_deserializer=repository__pb2.DownloadBentoRequest.FromString,
                    response_serializer=repository__pb2.DownloadBentoResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'bentoml.Yatai', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class Yatai(object):
    """Yatai RPC Server

    A stateful service that provides a complete BentoML model management
    and model serving/deployment workflow

    It provides two sets of APIs:
    Bento Repository: Manages saved Bento files, and making them available
    for serving in production environments
    Serving Deployment: Deploys saved Bento to a varity of different cloud
    platforms, track deploym//ent status, set up logging
    monitoring for your model serving workloads, and
    """

    @staticmethod
    def HealthCheck(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/bentoml.Yatai/HealthCheck',
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            yatai__service__pb2.HealthCheckResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetYataiServiceVersion(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/bentoml.Yatai/GetYataiServiceVersion',
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            yatai__service__pb2.GetYataiServiceVersionResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ApplyDeployment(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/bentoml.Yatai/ApplyDeployment',
            deployment__pb2.ApplyDeploymentRequest.SerializeToString,
            deployment__pb2.ApplyDeploymentResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def DeleteDeployment(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/bentoml.Yatai/DeleteDeployment',
            deployment__pb2.DeleteDeploymentRequest.SerializeToString,
            deployment__pb2.DeleteDeploymentResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetDeployment(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/bentoml.Yatai/GetDeployment',
            deployment__pb2.GetDeploymentRequest.SerializeToString,
            deployment__pb2.GetDeploymentResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def DescribeDeployment(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/bentoml.Yatai/DescribeDeployment',
            deployment__pb2.DescribeDeploymentRequest.SerializeToString,
            deployment__pb2.DescribeDeploymentResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ListDeployments(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/bentoml.Yatai/ListDeployments',
            deployment__pb2.ListDeploymentsRequest.SerializeToString,
            deployment__pb2.ListDeploymentsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def AddBento(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/bentoml.Yatai/AddBento',
            repository__pb2.AddBentoRequest.SerializeToString,
            repository__pb2.AddBentoResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def UpdateBento(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/bentoml.Yatai/UpdateBento',
            repository__pb2.UpdateBentoRequest.SerializeToString,
            repository__pb2.UpdateBentoResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetBento(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/bentoml.Yatai/GetBento',
            repository__pb2.GetBentoRequest.SerializeToString,
            repository__pb2.GetBentoResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def DangerouslyDeleteBento(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/bentoml.Yatai/DangerouslyDeleteBento',
            repository__pb2.DangerouslyDeleteBentoRequest.SerializeToString,
            repository__pb2.DangerouslyDeleteBentoResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ListBento(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/bentoml.Yatai/ListBento',
            repository__pb2.ListBentoRequest.SerializeToString,
            repository__pb2.ListBentoResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ContainerizeBento(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/bentoml.Yatai/ContainerizeBento',
            repository__pb2.ContainerizeBentoRequest.SerializeToString,
            repository__pb2.ContainerizeBentoResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def UploadBento(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_unary(request_iterator, target, '/bentoml.Yatai/UploadBento',
            repository__pb2.UploadBentoRequest.SerializeToString,
            repository__pb2.UploadBentoResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def DownloadBento(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/bentoml.Yatai/DownloadBento',
            repository__pb2.DownloadBentoRequest.SerializeToString,
            repository__pb2.DownloadBentoResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



AZURE_API_FUNCTION_JSON = """\
{{
  "scriptFile": "__init__.py",
  "bindings": [
    {{
      "authLevel": "{function_auth_level}",
      "type": "httpTrigger",
      "direction": "in",
      "name": "req",
      "methods": [
        "get",
        "post",
        "put",
        "delete",
        "patch"
      ],
      "route": "{{*route}}"
    }},
    {{
      "type": "http",
      "direction": "out",
      "name": "$return"
    }}
  ]
}}
"""
